# -*- coding:utf-8 -*-
# Author: Kei Choi(hanul93@gmail.com)

"""
InstallShield Archive Engine Plugin

This plugin handles InstallShield format for scanning and manipulation.
"""

import contextlib
import logging
import os
import struct
import zlib
from io import BytesIO

from kicomav.plugins import cryptolib
from kicomav.plugins import kavutil
from kicomav.kavcore import k2security
from kicomav.kavcore.plugin_base import ArchivePluginBase

# Module logger
logger = logging.getLogger(__name__)


# -------------------------------------------------------------------------
# InstallShield Functions
# -------------------------------------------------------------------------
# Constants
PREFER_BLOCK_SIZE = 4096 * 64
ISSIG = b"InstallShield"
ISSIG_strm = b"ISSetupStream"
_MAX_PATH = 260


def gen_key(seed: str) -> bytes:
    """Generate key from seed"""
    MAGIC_DEC = [0x13, 0x35, 0x86, 0x07]
    seed_bytes = seed.encode("utf-8")
    key = bytearray()
    for i, byte in enumerate(seed_bytes):
        key.append(byte ^ MAGIC_DEC[i % len(MAGIC_DEC)])
    return bytes(key)


def decode_byte(byte: int, key_byte: int) -> int:
    """Decode single byte"""
    # C code: return ~(key ^ (Byte * 16 | Byte >> 4));
    # Byte * 16 is left shift by 4, Byte >> 4 is right shift by 4
    rotated = ((byte << 4) & 0xFF) | (byte >> 4)
    xored = key_byte ^ rotated
    return (~xored) & 0xFF


def decode_data(data: bytes, offset: int, key: bytes) -> bytes:
    """Decode data with key"""
    result = bytearray(data)
    key_len = len(key)
    for i in range(len(result)):
        result[i] = decode_byte(result[i], key[(i + offset) % key_len])
    return bytes(result)


def decode_data_ustrm(data: bytes, offset: int, key: bytes) -> bytes:
    """Decode data for unicode stream (1024 byte blocks)"""
    result = bytearray()
    key_len = len(key)
    decoded_len = 0
    data_len = len(data)

    while decoded_len < data_len:
        decode_start = (decoded_len + offset) % 1024
        task_len = min(1024 - decode_start, data_len - decoded_len)

        # Decode this block
        block = data[decoded_len : decoded_len + task_len]
        decoded_block = decode_data(block, decode_start % key_len, key)
        result.extend(decoded_block)

        decoded_len += task_len

    return bytes(result)


def decode_file(
    fp_r,
    offset_r: int,
    length: int,
    encoded_block_len: int,
    key: bytes = None,
    decode_fun=None,
    need_decode: bool = False,
) -> (int, bytes):
    """Decode and return file content (optimized: list-append batching)"""
    if length <= 0:
        return 0, b""

    fp_r.seek(offset_r)

    len_read = min(PREFER_BLOCK_SIZE, encoded_block_len) if encoded_block_len else PREFER_BLOCK_SIZE
    offset_w = 0
    len_encoded_done = 0

    chunks = []  # use list to avoid O(n^2) bytes concatenation

    while length > 0:
        current_len_read = min(len_read, length)
        if encoded_block_len and encoded_block_len > length:
            encoded_block_len = length

        len_left = encoded_block_len if encoded_block_len else current_len_read
        len_encoded_done = 0

        while len_left > 0:
            chunk_size = min(len_read, len_left)
            pbuffer = fp_r.read(chunk_size)
            if not pbuffer:
                break

            if need_decode and key and decode_fun:
                pbuffer = decode_fun(pbuffer, len_encoded_done, key)

            chunks.append(pbuffer)

            offset_r += len(pbuffer)
            offset_w += len(pbuffer)
            len_encoded_done += len(pbuffer)
            len_left -= len(pbuffer)

        if encoded_block_len:
            length -= encoded_block_len
        else:
            length -= current_len_read

    return offset_r, b"".join(chunks)


def get_is_file_attributes(fp, data_offset: int, is_hdr_type: int):
    """Get IS file attributes (standard format)"""
    fp.seek(data_offset)
    data = fp.read(ISFileAttributes.SIZE)

    if len(data) != ISFileAttributes.SIZE:
        return data_offset, None, ""

    attr = ISFileAttributes(data)
    return data_offset + ISFileAttributes.SIZE, attr, ""


def get_is_file_attributes_ustrm(fp, data_offset: int, is_hdr_type: int):
    """Get IS file attributes (unicode stream format)"""
    fp.seek(data_offset)
    data = fp.read(ISFileAttributesX.SIZE)

    if len(data) != ISFileAttributesX.SIZE:
        return data_offset, None, ""

    attr_x = ISFileAttributesX(data)

    # Validate filename length
    if attr_x.filename_len <= 0 or attr_x.filename_len >= _MAX_PATH * 2:
        return data_offset, None, ""

    # Skip extra data for type 4
    if is_hdr_type == 4:
        fp.seek(ISFileAttributesX.SIZE, 1)

    # Read unicode filename
    filename_data = fp.read(attr_x.filename_len)
    if len(filename_data) != attr_x.filename_len:
        return data_offset, None, ""

    try:
        # Decode UTF-16LE filename
        filename = filename_data.decode("utf-16le").rstrip("\x00")
        attr_x.file_name = filename
    except:
        attr_x.file_name = f"0x{data_offset:X}"

    # Also store seed for decryption (UTF-8 encoded filename)
    try:
        g_Seed = filename_data.decode("utf-16le").rstrip("\x00")
    except:
        g_Seed = ""

    new_offset = fp.tell()

    return new_offset, attr_x, g_Seed


def rol1(byte, count):
    """Rotate left 1 byte"""
    return ((byte << count) | (byte >> (8 - count))) & 0xFF


def decrypt(encrypted_data, key: bytes) -> bytes:
    decrypted = bytearray(len(encrypted_data))
    n = 0
    key_len = len(key)

    for i in range(len(encrypted_data)):
        v52 = encrypted_data[i]
        v52 = rol1(v52, 3)
        decrypted[i] = v52 ^ key[n]

        n += 1
        if n >= key_len:
            n = 0

    return bytes(decrypted)


def decode_exe(fp, key: bytes, exe_start_off: int, exe_end_off: int) -> bytes:
    """Decode embedded setup.exe"""
    fp.seek(exe_start_off)
    encrypted_exe_data = fp.read(0x300)

    if len(encrypted_exe_data) < 0x300:
        return b""

    return decrypt(encrypted_exe_data, key[:16])


def inflate_file(compressed_data: bytes) -> (bool, bytes):
    """Inflate (decompress) a zlib compressed file"""
    # Check if data actually looks compressed (has zlib header)
    # zlib header usually starts with 0x78 (common: 0x789C, 0x78DA, 0x785E, 0x7801)
    if len(compressed_data) < 2:
        return False, None

    try:
        decompressed_data = zlib.decompress(compressed_data)
    except zlib.error:
        return False, None

    return True, decompressed_data


class ISHeader:
    """InstallShield Archive Header - 46 bytes (0x2E)"""

    SIZE = 46

    def __init__(self, data: bytes):
        # char SIG[14], uint16_t num_files, uint32_t type, uint8_t x4[8], uint16_t x5, uint8_t x6[16]
        fields = struct.unpack("<14sHI8sH16s", data[:46])
        self.sig = fields[0].rstrip(b"\x00")
        self.num_files = fields[1]
        self.type = fields[2]
        self.x4 = fields[3]
        self.x5 = fields[4]
        self.x6 = fields[5]


class ISFileAttributes:
    """IS File Attributes - 312 bytes (0x138)"""

    SIZE = 312

    def __init__(self, data: bytes):
        # char file_name[260], uint32_t encoded_flags, uint32_t x3, uint32_t file_len,
        # uint8_t x5[8], uint16_t is_unicode_launcher, uint8_t x7[30]
        fields = struct.unpack("<260sIII8sH30s", data[:312])
        self.file_name = fields[0].rstrip(b"\x00").decode("utf-8", errors="ignore")
        self.encoded_flags = fields[1]
        self.x3 = fields[2]
        self.file_len = fields[3]
        self.x5 = fields[4]
        self.is_unicode_launcher = fields[5]
        self.x7 = fields[6]


class ISFileAttributesX:
    """IS File Attributes X - 24 bytes (0x18) for ISSetupStream"""

    SIZE = 24

    def __init__(self, data: bytes):
        # uint32_t filename_len, uint32_t encoded_flags, uint8_t x3[2], uint32_t file_len,
        # uint8_t x5[8], uint16_t is_unicode_launcher
        fields = struct.unpack("<II2sI8sH", data[:24])
        self.filename_len = fields[0]
        self.encoded_flags = fields[1]
        self.x3 = fields[2]
        self.file_len = fields[3]
        self.x5 = fields[4]
        self.is_unicode_launcher = fields[5]
        self.file_name = ""


class InstallShield:
    def __init__(self, fname, data_off: int):
        self.fname = fname
        self.data_off = data_off
        self.fp = None
        self.header = None
        self.fsize = 0
        self.num_files = 0
        self.install_name = []
        self.g_Seed = ""

    def __del__(self):
        if self.fp:
            self.close()

    def close(self):
        if self.fp:
            self.fp.close()
            self.fp = None

    def parse(self):
        with contextlib.suppress(IOError, OSError, ValueError):
            self.fp = open(self.fname, "rb")
            self.fp.seek(self.data_off)

            data = self.fp.read(ISHeader.SIZE)
            is_hdr = ISHeader(data)

            # Check signature
            if is_hdr.sig == ISSIG or is_hdr.sig == ISSIG_strm:
                # Check type
                if is_hdr.type > 4:
                    return False

                offset = self.data_off + ISHeader.SIZE
                self.data_off = offset

                # Determine format
                self.is_stream = is_hdr.sig == ISSIG_strm
                self.encoded_block_len = 0x4000 if self.is_stream else PREFER_BLOCK_SIZE
                get_attr_fun = get_is_file_attributes_ustrm if self.is_stream else get_is_file_attributes
                self.decode_fun = decode_data_ustrm if self.is_stream else decode_data

                self.num_files = is_hdr.num_files
                # print(f"Files total: {self.num_files}")
                # print("Extracting:")

                self.is_hdr = is_hdr

                for i in range(self.num_files):
                    # Get file attributes
                    offset, file_attr, self.g_Seed = get_attr_fun(self.fp, self.data_off, self.is_hdr.type)

                    if offset <= self.data_off or file_attr is None:
                        break

                    self.data_off = offset
                    file_data_offset = offset
                    self.data_off += file_attr.file_len

                    self.install_name.append(
                        (
                            file_data_offset,
                            file_attr.file_len,
                            file_attr.file_name,
                            file_attr,
                            self.data_off,
                            "HEADER_FNAME",
                        )
                    )

            else:
                # Check CAB signature
                header_size = kavutil.get_uint32(data, 0)
                cab_off = self.data_off + header_size + 4
                self.fp.seek(cab_off)
                cab_sig = self.fp.read(0x10)

                if cab_sig[0:4] == b"MSCF":
                    cab_size = kavutil.get_uint32(cab_sig, 8)

                    self.install_name.append(
                        (
                            cab_off,
                            cab_size,
                            "InstallShield_Data.cab",
                            None,
                            None,
                            "HEADER_CAB",
                        )
                    )

            # Check exist of setup.exe
            self.fp.seek(-0x30, 2)
            tail_header = self.fp.read(0x30)
            exe_start_off = kavutil.get_uint32(tail_header, 0xC)
            exe_end_off = kavutil.get_uint32(tail_header, 0x1C)

            decrypted_exe_header = decode_exe(self.fp, tail_header, exe_start_off, exe_end_off)
            if decrypted_exe_header[:2] == b"MZ":
                self.install_name.append(
                    (
                        exe_start_off,
                        exe_end_off - exe_start_off,
                        "InstallShield_Setup.exe",
                        None,
                        decrypted_exe_header,
                        "ENCRYPT_EXE",
                    )
                )

            return True

        return False

    def namelist(self):
        flist = []

        flist.extend((f[2] for f in self.install_name))
        return flist

    def read(self, fname):
        for f in self.install_name:
            if f[2] == fname and self.fp:
                f_type = f[5]

                if f_type == "HEADER_FNAME":
                    file_data_offset, _, _, file_attr, data_offset, _ = f

                    # Prepare decoder
                    is_need_inflate = False
                    # seed = self.g_Seed if self.is_stream else file_attr.file_name
                    seed = file_attr.file_name
                    key = gen_key(seed)

                    # Determine encoding type
                    has_type_2_or_4 = file_attr.encoded_flags & 6
                    has_type_4 = file_attr.encoded_flags & 4
                    encoded_block_len_i = self.encoded_block_len
                    need_decode = False

                    if has_type_4 and has_type_2_or_4:
                        encoded_block_len_i = 1024
                        need_decode = True

                    offset, return_data = decode_file(
                        self.fp,
                        file_data_offset,
                        file_attr.file_len,
                        encoded_block_len_i,
                        key,
                        self.decode_fun,
                        need_decode,
                    )

                    if offset != data_offset:
                        self.fp.seek(data_offset)
                    else:
                        n_2trans = 1

                        # Second pass for certain types
                        if n_2trans and not has_type_4 and has_type_2_or_4:
                            # Convert to bytesio
                            fp_w = BytesIO(return_data)
                            fp_w.seek(0)

                            decode_fun_2 = decode_data if self.is_stream else self.decode_fun
                            offset, return_data = decode_file(
                                fp_w, 0, file_attr.file_len, file_attr.file_len, key, decode_fun_2, True
                            )

                            if offset != file_attr.file_len:
                                self.fp.seek(data_offset)

                            fp_w.close()

                        # Check if inflation needed
                        if file_attr.is_unicode_launcher:
                            is_need_inflate = True

                    # Inflate if needed
                    if is_need_inflate:
                        _, return_data = inflate_file(return_data)

                    return return_data

                elif f_type == "ENCRYPT_EXE":
                    exe_start_off, exe_len, _, _, decrypted_exe_header, _ = f
                    self.fp.seek(exe_start_off)
                    exe_data = self.fp.read(exe_len)
                    return decrypted_exe_header + exe_data[0x300:]

                elif f_type == "HEADER_CAB":
                    cab_off, cab_len, _, _, _, _ = f
                    self.fp.seek(cab_off)
                    cab_data = self.fp.read(cab_len)
                    return cab_data

        return None


class InstallShieldSetupEmbeddedFile:
    def __init__(self, filename, data_off=0):
        self.filename = filename
        self.data_off = data_off
        self.fp = None
        self.key = b"%eR@toPm|<#YKs$^"
        self.install_name = []
        self.CHUNK_SIZE = 1024

    def __del__(self):
        if self.fp:
            self.close()

    def close(self):
        if self.fp:
            self.fp.close()
            self.fp = None

    def parse(self):
        with contextlib.suppress(IOError, OSError, ValueError):
            self.fp = open(self.filename, "rb")
            self.fp.seek(self.data_off)
            data = self.fp.read()

            off = len(data) - 712
            n = 0

            while off >= 0:
                buf = data[off : off + 712]
                decrypted = cryptolib.ishield_decrypt(buf, self.key)

                key_size = kavutil.get_uint32(decrypted, 668)
                v2 = kavutil.get_uint32(decrypted, 12)
                file_name_size = kavutil.get_uint32(decrypted, 144)
                crc32_value = kavutil.get_uint32(decrypted, 704)
                file_size = kavutil.get_uint32(decrypted, 708)

                if key_size <= 0x20 and v2 <= 0x40 and file_name_size <= 0x104:
                    decrypted = cryptolib.reset_padding(decrypted, 668, key_size, 0x20)
                    decrypted = cryptolib.reset_padding(decrypted, 12, v2, 0x40)
                    decrypted = cryptolib.reset_padding(decrypted, 144, file_name_size, 0x104)
                    crc32_value_calculated = int(zlib.crc32(decrypted[:704]) & 0xFFFFFFFF)
                    if crc32_value == crc32_value_calculated and file_name_size != 0:
                        file_name = decrypted[148 : 148 + file_name_size].decode("utf-16le", errors="ignore")
                        file_name = os.path.expandvars(file_name)
                        file_name = os.path.basename(file_name)

                        dec_key = decrypted[668 + 4 : 668 + 4 + key_size]

                        self.install_name.append(
                            (
                                off - file_size,
                                file_size,
                                file_name,
                                dec_key,
                                None,
                                "EMBED_EXE",
                            )
                        )
                    off -= 712 + file_size
                else:
                    break

            return True
        return False

    def namelist(self):
        flist = []

        flist.extend((f[2] for f in self.install_name))
        return flist

    def read(self, fname):
        for f in self.install_name:
            offset, compressed_size, file_name, dec_key, _, _ = f
            decrypted_chunks = []

            if file_name != fname and self.fp:
                continue

            self.fp.seek(offset)

            remaining = compressed_size

            while remaining > 0:
                chunk_size = min(self.CHUNK_SIZE, remaining)
                chunk = self.fp.read(chunk_size)

                if len(chunk) != chunk_size:
                    break

                decrypted_chunk = decrypt(chunk, dec_key)
                decrypted_chunks.append(decrypted_chunk)

                remaining -= chunk_size

            decrypted_data = b"".join(decrypted_chunks)
            _, decrypted_data = inflate_file(decrypted_data)

            return decrypted_data

        return None


# -------------------------------------------------------------------------
# KavMain class
# -------------------------------------------------------------------------
class KavMain(ArchivePluginBase):
    """InstallShield archive handler plugin.

    This plugin provides functionality for:
    - Detecting InstallShield format
    - Listing files within InstallShield installers
    - Extracting files from InstallShield installers
    """

    def __init__(self):
        """Initialize the InstallShield plugin."""
        super().__init__(
            author="Kei Choi",
            version="1.0",
            title="InstallShield Engine",
            kmd_name="ishield",
        )

    def __get_handle(self, filename, data_off=0):
        """Get or create handle for InstallShield file.

        Args:
            filename: Path to InstallShield file
            data_off: Data offset

        Returns:
            InstallShield object or None
        """
        if filename in self.handle:
            return self.handle.get(filename, None)

        try:
            zfile = InstallShield(filename, data_off)
            self.handle[filename] = zfile
            return zfile

        except (IOError, OSError) as e:
            logger.debug("Failed to open InstallShield file %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error opening InstallShield file %s: %s", filename, e)

        return None

    def __get_handle_setup_embedded_file(self, filename, data_off=0):
        """Get or create handle for InstallShield embedded file.

        Args:
            filename: Path to file
            data_off: Data offset

        Returns:
            InstallShieldSetupEmbeddedFile object or None
        """
        if filename in self.handle:
            return self.handle.get(filename, None)

        try:
            zfile = InstallShieldSetupEmbeddedFile(filename, data_off)
            self.handle[filename] = zfile
            return zfile

        except (IOError, OSError) as e:
            logger.debug("Failed to open InstallShield embedded file %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error opening InstallShield embedded file %s: %s", filename, e)

        return None

    def arclist(self, filename, fileformat, password=None):
        """List files in the archive.

        Args:
            filename: Path to archive file
            fileformat: Format info from format() method

        Returns:
            List of [engine_id, filename] pairs
        """
        file_scan_list = []

        try:
            if "ff_installshield" in fileformat:
                zfile = self.__get_handle(filename, fileformat["ff_installshield"]["Attached_Pos"])
                if zfile and zfile.parse():
                    for name in zfile.namelist():
                        # CWE-22: Path traversal prevention
                        if k2security.is_safe_archive_member(name):
                            file_scan_list.append(["arc_installshield", name])

            elif "ff_installshield_setup_stream" in fileformat:
                zfile = self.__get_handle(filename, fileformat["ff_installshield_setup_stream"]["Attached_Pos"])
                if zfile and zfile.parse():
                    for name in zfile.namelist():
                        # CWE-22: Path traversal prevention
                        if k2security.is_safe_archive_member(name):
                            file_scan_list.append(["arc_installshield", name])

            elif "ff_installshield_embedded_file" in fileformat:
                zfile = self.__get_handle_setup_embedded_file(filename)
                if zfile and zfile.parse():
                    for name in zfile.namelist():
                        # CWE-22: Path traversal prevention
                        if k2security.is_safe_archive_member(name):
                            file_scan_list.append(["arc_installshield_embedded_file", name])

        except (IOError, OSError) as e:
            logger.debug("Archive list IO error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error listing archive %s: %s", filename, e)

        return file_scan_list

    def unarc(self, arc_engine_id, arc_name, fname_in_arc):
        """Extract a file from the archive.

        Args:
            arc_engine_id: Engine ID ('arc_installshield', 'arc_installshield_embedded_file')
            arc_name: Path to archive file
            fname_in_arc: Name of file to extract

        Returns:
            Extracted file data, or None on error
        """
        # CWE-22: Path traversal prevention
        if not k2security.is_safe_archive_member(fname_in_arc):
            logger.warning("Unsafe archive member rejected: %s in %s", fname_in_arc, arc_name)
            return None

        try:
            if arc_engine_id == "arc_installshield":
                zfile = self.__get_handle(arc_name)
                if zfile is None:
                    return None
                return zfile.read(fname_in_arc)

            elif arc_engine_id == "arc_installshield_embedded_file":
                zfile = self.__get_handle_setup_embedded_file(arc_name)
                if zfile is None:
                    return None
                return zfile.read(fname_in_arc)

        except (IOError, OSError) as e:
            logger.debug("Archive extract IO error for %s in %s: %s", fname_in_arc, arc_name, e)
        except Exception as e:
            logger.warning("Unexpected error extracting %s from %s: %s", fname_in_arc, arc_name, e)

        return None

    def arcclose(self):
        """Close all open archive handles."""
        for fname in list(self.handle.keys()):
            try:
                zfile = self.handle.get(fname)
                if zfile:
                    zfile.close()
            except (IOError, OSError) as e:
                logger.debug("Archive close IO error for %s: %s", fname, e)
            except Exception as e:
                logger.debug("Archive close error for %s: %s", fname, e)
            finally:
                self.handle.pop(fname, None)



# -*- coding:utf-8 -*-
# Author: Kei Choi(hanul93@gmail.com)

"""
EGG Archive Engine Plugin

This plugin handles EGG archive format for scanning and manipulation.
"""

import bz2
import contextlib
import mmap
import os
import shutil
import zipfile
import zlib
import logging

from kicomav.plugins import kernel
from kicomav.plugins import kavutil
from kicomav.kavcore import k2security
from kicomav.kavcore.plugin_base import ArchivePluginBase

# Module logger
logger = logging.getLogger(__name__)


# -------------------------------------------------------------------------
# EGG constants
# -------------------------------------------------------------------------
SIZE_EGG_HEADER = 14

COMPRESS_METHOD_STORE = 0
COMPRESS_METHOD_DEFLATE = 1
COMPRESS_METHOD_BZIP2 = 2
COMPRESS_METHOD_AZO = 3
COMPRESS_METHOD_LZMA = 4


# -------------------------------------------------------------------------
# class EggFile
# -------------------------------------------------------------------------
class EggFile:
    """EGG file handler."""

    def __init__(self, filename):
        self.fp = None
        self.mm = None
        self.data_size = 0
        self.egg_pos = None

        with contextlib.suppress(IOError):
            self.data_size = os.path.getsize(filename)
            self.fp = open(filename, "rb")
            self.mm = mmap.mmap(self.fp.fileno(), 0, access=mmap.ACCESS_READ)

    def close(self):
        if self.mm:
            self.mm.close()
            self.mm = None

        if self.fp:
            self.fp.close()
            self.fp = None

    def read(self, filename):
        ret_data = None

        with contextlib.suppress(Exception):
            fname = self.__FindFirstFileName__()
            while fname:
                if fname == filename:
                    data, method, self.egg_pos = self.__ReadBlockData__()
                    if method == COMPRESS_METHOD_STORE:
                        ret_data = data
                        break
                    elif method == COMPRESS_METHOD_DEFLATE:
                        ret_data = zlib.decompress(data, -15)
                        break
                    elif method == COMPRESS_METHOD_BZIP2:
                        ret_data = bz2.decompress(data)
                        break
                fname = self.__FindNextFileName__()

        return ret_data

    def namelist(self):
        name_list = []

        with contextlib.suppress(Exception):
            fname = self.__FindFirstFileName__()
            while fname:
                name_list.append(fname)
                fname = self.__FindNextFileName__()

        return name_list

    def __FindFirstFileName__(self):
        self.egg_pos = 0
        fname, self.egg_pos = self.__GetFileName__(self.egg_pos)
        return fname

    def __FindNextFileName__(self):
        fname, self.egg_pos = self.__GetFileName__(self.egg_pos)
        return fname

    def __GetFileName__(self, egg_pos):
        mm = self.mm
        data_size = self.data_size

        with contextlib.suppress(SystemError):
            while egg_pos < data_size:
                magic = kavutil.get_uint32(mm, egg_pos)

                if magic == 0x0A8591AC:  # Filename Header
                    size, fname = self.__EGG_Filename_Header__(mm[egg_pos:])
                    if size == -1:
                        raise SystemError
                    egg_pos += size
                    return fname, egg_pos
                else:
                    egg_pos = self.__DefaultMagicIDProc__(magic, egg_pos)
                    if egg_pos == -1:
                        raise SystemError

        return None, -1

    def __ReadBlockData__(self):
        egg_pos = self.egg_pos
        mm = self.mm
        data_size = self.data_size

        with contextlib.suppress(SystemError):
            while egg_pos < data_size:
                magic = kavutil.get_uint32(mm, egg_pos)

                if magic == 0x02B50C13:  # Block Header
                    size = self.__EGG_Block_Header_Size__(mm[egg_pos:])
                    if size == -1:
                        raise SystemError

                    compress_method = int(mm[egg_pos + 4])
                    compress_size = kavutil.get_uint32(mm, egg_pos + 10)
                    compressed_data = mm[egg_pos + 22 : egg_pos + 22 + compress_size]
                    egg_pos += size
                    return compressed_data, compress_method, egg_pos
                else:
                    egg_pos = self.__DefaultMagicIDProc__(magic, egg_pos)
                    if egg_pos == -1:
                        raise SystemError

        return None, -1, -1

    def __DefaultMagicIDProc__(self, magic, egg_pos):
        mm = self.mm
        data_size = self.data_size

        try:
            if egg_pos < data_size:
                if magic == 0x41474745:  # EGG Header
                    if self.__EGG_Header__(mm) == -1:
                        raise SystemError
                    egg_pos += SIZE_EGG_HEADER
                elif magic == 0x0A8590E3:  # File Header
                    egg_pos += 16
                elif magic == 0x02B50C13:  # Block Header
                    size = self.__EGG_Block_Header_Size__(mm[egg_pos:])
                    if size == -1:
                        raise SystemError
                    egg_pos += size
                elif magic == 0x08D1470F:  # Encrypt Header
                    size = self.__EGG_Encrypt_Header_Size__(mm[egg_pos:])
                    if size == -1:
                        raise SystemError
                    egg_pos += size
                elif magic == 0x2C86950B:  # Windows File Information
                    egg_pos += 16
                elif magic == 0x1EE922E5:  # Posix File Information
                    egg_pos += 27
                elif magic == 0x07463307:  # Dummy Header
                    size = self.__EGG_Dummy_Header_Size__(mm[egg_pos:])
                    if size == -1:
                        raise SystemError
                    egg_pos += size
                elif magic == 0x0A8591AC:  # Filename Header
                    size, fname = self.__EGG_Filename_Header__(mm[egg_pos:])
                    if size == -1:
                        raise SystemError
                    egg_pos += size
                elif magic == 0x04C63672:  # Comment Header
                    raise SystemError  # Not supported
                elif magic == 0x24F5A262:  # Split Compression
                    egg_pos += 15
                elif magic == 0x24E5A060:  # Solid Compression
                    egg_pos += 7
                elif magic == 0x08E28222:  # End of File Header
                    egg_pos += 4
                else:
                    raise SystemError
        except SystemError:
            return -1

        return egg_pos

    def __EGG_Header__(self, data):
        with contextlib.suppress(SystemError):
            magic = kavutil.get_uint32(data, 0)
            if magic != 0x41474745:
                raise SystemError

            version = kavutil.get_uint16(data, 4)
            if version != 0x0100:
                raise SystemError

            header_id = kavutil.get_uint32(data, 6)
            if header_id == 0:
                raise SystemError

            reserved = kavutil.get_uint32(data, 10)
            if reserved != 0:
                raise SystemError

            return 0

        return -1

    def __EGG_Encrypt_Header_Size__(self, data):
        with contextlib.suppress(SystemError):
            encrypt_method = int(data[7])
            if encrypt_method == 0:
                return 24
            elif encrypt_method == 1:
                return 28
            elif encrypt_method == 2:
                return 36
            else:
                raise SystemError

        return -1

    def __EGG_Dummy_Header_Size__(self, data):
        with contextlib.suppress(Exception):
            dummy_size = kavutil.get_uint16(data, 5)
            return 7 + dummy_size

        return -1

    def __EGG_Filename_Header__(self, data):
        size = -1
        fname = None

        with contextlib.suppress(Exception):
            fname_size = kavutil.get_uint16(data, 5)
            fname = data[7 : 7 + fname_size]
            size = 7 + fname_size

        return size, fname.decode("utf-8") if fname else None

    def __EGG_Block_Header_Size__(self, data):
        size = -1

        with contextlib.suppress(Exception):
            block_size = 18 + 4
            compress_size = kavutil.get_uint32(data, 10)
            size = block_size + compress_size

        return size


# -------------------------------------------------------------------------
# class KavMain
# -------------------------------------------------------------------------
class KavMain(ArchivePluginBase):
    """EGG archive handler plugin.

    This plugin provides functionality for:
    - Detecting EGG archive format
    - Listing files within archives
    - Extracting files from archives
    - Converting EGG to ZIP format
    """

    def __init__(self):
        """Initialize the EGG plugin."""
        super().__init__(
            author="Kei Choi",
            version="1.1",
            title="Egg Archive Engine",
            kmd_name="egg",
        )

    def getinfo(self):
        """Get plugin information.

        Returns:
            Dictionary containing plugin metadata
        """
        info = super().getinfo()
        info["engine_type"] = kernel.ARCHIVE_ENGINE
        info["make_arc_type"] = kernel.MASTER_PACK
        return info

    def __get_handle(self, filename):
        """Get or create handle for EGG file.

        Args:
            filename: Path to EGG file

        Returns:
            EggFile object
        """
        if filename in self.handle:
            return self.handle.get(filename, None)
        else:
            zfile = EggFile(filename)
            self.handle[filename] = zfile
            return zfile

    def format(self, filehandle, filename, filename_ex):
        """Analyze and detect EGG format.

        Args:
            filehandle: File data (memory mapped)
            filename: Path to archive file
            filename_ex: Extended filename info

        Returns:
            Dictionary with format info, or None if not recognized
        """
        try:
            if filehandle[:4] == b"EGGA":
                return {"ff_egg": "EGG"}
        except (IOError, OSError) as e:
            logger.debug("Format detection IO error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error in format detection for %s: %s", filename, e)

        return None

    def arclist(self, filename, fileformat, password=None):
        """List files in the archive.

        Args:
            filename: Path to archive file
            fileformat: Format info from format() method

        Returns:
            List of [engine_id, filename] pairs
        """
        file_scan_list = []

        try:
            if "ff_egg" in fileformat:
                zfile = self.__get_handle(filename)
                for name in zfile.namelist():
                    # CWE-22: Path traversal prevention
                    if k2security.is_safe_archive_member(name):
                        file_scan_list.append(["arc_egg", name])

        except (IOError, OSError) as e:
            logger.debug("Archive list IO error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error listing archive %s: %s", filename, e)

        return file_scan_list

    def unarc(self, arc_engine_id, arc_name, fname_in_arc):
        """Extract a file from the archive.

        Args:
            arc_engine_id: Engine ID ('arc_egg')
            arc_name: Path to archive file
            fname_in_arc: Name of file to extract

        Returns:
            Extracted file data, or None on error
        """
        # CWE-22: Path traversal prevention
        if not k2security.is_safe_archive_member(fname_in_arc):
            logger.warning("Unsafe archive member rejected: %s in %s", fname_in_arc, arc_name)
            return None

        try:
            if arc_engine_id == "arc_egg":
                zfile = self.__get_handle(arc_name)
                return zfile.read(fname_in_arc)

        except (IOError, OSError) as e:
            logger.debug("Archive extract IO error for %s in %s: %s", fname_in_arc, arc_name, e)
        except Exception as e:
            logger.warning("Unexpected error extracting %s from %s: %s", fname_in_arc, arc_name, e)

        return None

    def arcclose(self):
        """Close all open archive handles."""
        for fname in list(self.handle.keys()):
            try:
                zfile = self.handle[fname]
                zfile.close()
            except (IOError, OSError) as e:
                logger.debug("Archive close IO error for %s: %s", fname, e)
            except Exception as e:
                logger.debug("Archive close error for %s: %s", fname, e)
            finally:
                self.handle.pop(fname, None)

    def mkarc(self, arc_engine_id, arc_name, file_infos):
        """Create archive (converts EGG to ZIP format).

        Args:
            arc_engine_id: Engine ID ('arc_egg')
            arc_name: Path to archive file
            file_infos: List of file info structures

        Returns:
            True if successful, False otherwise
        """
        if arc_engine_id != "arc_egg":
            return False

        try:
            uname = kavutil.uniq_string()
            tname = f".tmp_{uname}"
            oname = f".org_{uname}"

            # Check if already converted to ZIP
            with open(arc_name, "rb") as f:
                if f.read(2) != b"PK":
                    self._convert_egg_to_zip(arc_name, tname)

            shutil.move(arc_name, arc_name + oname)
            shutil.move(arc_name + tname, arc_name)

            # Add cleaned file to converted ZIP
            kavutil.make_zip(arc_name, file_infos)

            # Clean up original file
            fname = arc_name + oname
            if os.path.exists(fname):
                with contextlib.suppress(k2security.SecurityError):
                    fname_dir = os.path.dirname(fname) or os.getcwd()
                    k2security.safe_remove_file(fname, fname_dir)

            return True

        except (IOError, OSError) as e:
            logger.error("Archive creation IO error for %s: %s", arc_name, e)
        except Exception as e:
            logger.error("Unexpected error creating archive %s: %s", arc_name, e)

        return False

    def _convert_egg_to_zip(self, arc_name, tname):
        """Convert EGG to ZIP format.

        Args:
            arc_name: Path to EGG file
            tname: Temporary file suffix
        """
        zfile_r = None
        try:
            with zipfile.ZipFile(arc_name + tname, "w") as zfile_w:
                zfile_r = EggFile(arc_name)
                for name in zfile_r.namelist():
                    data = zfile_r.read(name)
                    zfile_w.writestr(name, data, compress_type=zipfile.ZIP_DEFLATED)
        finally:
            if zfile_r is not None:
                zfile_r.close()



# -*- coding:utf-8 -*-
# Author: Kei Choi(hanul93@gmail.com)

"""
CAB Archive Engine Plugin

This plugin handles CAB archive format for scanning and manipulation.
"""

import contextlib
import os
import tempfile
import shutil
import logging

from kicomav.plugins import kernel
from kicomav.kavcore import k2security
from kicomav.kavcore.plugin_base import ArchivePluginBase

# Module logger
logger = logging.getLogger(__name__)

try:
    from pycabfile import CabFile

    PYCABFILE_AVAILABLE = True
except ImportError:
    PYCABFILE_AVAILABLE = False


# -------------------------------------------------------------------------
# class KavMain
# -------------------------------------------------------------------------
class KavMain(ArchivePluginBase):
    """CAB archive handler plugin.

    This plugin provides functionality for:
    - Detecting CAB archive format
    - Listing files within archives
    - Extracting files from archives
    - Creating/updating archives
    """

    def __init__(self):
        """Initialize the CAB plugin."""
        super().__init__(
            author="Kei Choi",
            version="3.0",
            title="CAB Archive Engine",
            kmd_name="cab",
        )
        self.temp_path = {}
        self.root_temp_path = None

    def _custom_init(self) -> int:
        """Custom initialization for CAB plugin.

        Returns:
            0 for success, -1 for failure
        """
        if not PYCABFILE_AVAILABLE:
            if self.verbose:
                logger.info("pycabfile package is not available")
            return -1

        pid = os.getpid()
        self.root_temp_path = os.path.join(tempfile.gettempdir(), "ktmp_cab_%05x" % pid)
        return 0

    def _custom_uninit(self) -> int:
        """Custom cleanup for CAB plugin.

        Returns:
            0 for success
        """
        self.arcclose()
        return 0

    def getinfo(self):
        """Get plugin information.

        Returns:
            Dictionary containing plugin metadata
        """
        info = super().getinfo()
        info["engine_type"] = kernel.ARCHIVE_ENGINE
        return info

    def __get_handle(self, filename):
        """Get or create handle for CAB file.

        Args:
            filename: Path to CAB file

        Returns:
            CabFile object or None
        """
        if filename in self.handle:
            return self.handle[filename]

        try:
            cab_file = CabFile(filename, "r")
            self.handle[filename] = cab_file

            if self.root_temp_path and not os.path.exists(self.root_temp_path):
                os.makedirs(self.root_temp_path, exist_ok=True)

            self.temp_path[filename] = tempfile.mkdtemp(prefix="ktmp", dir=self.root_temp_path)
            return cab_file

        except (IOError, OSError) as e:
            logger.debug("Failed to open CAB file %s: %s", filename, e)
        except Exception as e:
            logger.debug("Error opening CAB file %s: %s", filename, e)

        return None

    def format(self, filehandle, filename, filename_ex):
        """Analyze and detect CAB format.

        Args:
            filehandle: File data (memory mapped)
            filename: Path to archive file
            filename_ex: Extended filename info

        Returns:
            Dictionary with format info, or None if not recognized
        """
        if not PYCABFILE_AVAILABLE:
            return None

        try:
            if len(filehandle) >= 4 and filehandle[:4] == b"MSCF":
                with CabFile(filename, "r") as cab_file:
                    file_list = cab_file.namelist()

                fileformat = {
                    "size": len(filehandle),
                    "file_count": len(file_list),
                }
                return {"ff_cab": fileformat}

        except (IOError, OSError) as e:
            logger.debug("Format detection IO error for %s: %s", filename, e)
        except Exception as e:
            logger.debug("Format detection error for %s: %s", filename, e)

        return None

    def arclist(self, filename, fileformat, password=None):
        """List files in the archive.

        Args:
            filename: Path to archive file
            fileformat: Format info from format() method

        Returns:
            List of [engine_id, filename] pairs
        """
        file_scan_list = []

        if "ff_cab" not in fileformat:
            return file_scan_list

        try:
            cab_file = self.__get_handle(filename)
            if cab_file is None:
                return file_scan_list

            file_list = cab_file.namelist()

            for filename_in_cab in file_list:
                # CWE-22: Path traversal prevention
                if k2security.is_safe_archive_member(filename_in_cab):
                    file_scan_list.append(["arc_cab", filename_in_cab])

        except (IOError, OSError) as e:
            logger.debug("Archive list IO error for %s: %s", filename, e)
        except Exception as e:
            logger.debug("Archive list error for %s: %s", filename, e)

        return file_scan_list

    def unarc(self, arc_engine_id, arc_name, fname_in_arc):
        """Extract a file from the archive.

        Args:
            arc_engine_id: Engine ID ('arc_cab')
            arc_name: Path to archive file
            fname_in_arc: Name of file to extract

        Returns:
            Extracted file data, or None on error
        """
        # CWE-22: Path traversal prevention
        if not k2security.is_safe_archive_member(fname_in_arc):
            logger.warning("Unsafe archive member rejected: %s in %s", fname_in_arc, arc_name)
            return None

        if arc_engine_id != "arc_cab":
            return None

        try:
            cab_file = self.handle.get(arc_name)
            if cab_file is None:
                return None

            file_data = cab_file.read(fname_in_arc)
            if file_data:
                return file_data

        except KeyError:
            logger.debug("File %s not found in CAB archive", fname_in_arc)
        except (IOError, OSError) as e:
            logger.debug("Archive extract IO error for %s in %s: %s", fname_in_arc, arc_name, e)
        except Exception as e:
            logger.debug("Archive extract error for %s in %s: %s", fname_in_arc, arc_name, e)

        return None

    def arcclose(self):
        """Close all open archive handles."""
        for fname in list(self.handle.keys()):
            try:
                cab_file = self.handle.get(fname)
                if cab_file:
                    cab_file.close()

                # Delete temporary directory
                temp_path = self.temp_path.get(fname)
                if temp_path and os.path.exists(temp_path):
                    with contextlib.suppress(OSError):
                        if os.path.isdir(temp_path):
                            shutil.rmtree(temp_path)
                        else:
                            os.remove(temp_path)

            except (IOError, OSError) as e:
                logger.debug("Archive close IO error for %s: %s", fname, e)
            except Exception as e:
                logger.debug("Archive close error for %s: %s", fname, e)
            finally:
                self.handle.pop(fname, None)
                self.temp_path.pop(fname, None)

        # Delete root temporary directory
        if self.root_temp_path and os.path.exists(self.root_temp_path):
            with contextlib.suppress(OSError):
                shutil.rmtree(self.root_temp_path)

    def mkarc(self, arc_engine_id, arc_name, file_infos):
        """Create a CAB archive.

        Args:
            arc_engine_id: Engine ID ('arc_cab')
            arc_name: Path to archive file
            file_infos: List of file info structures

        Returns:
            True if successful, False otherwise
        """
        if arc_engine_id != "arc_cab":
            return False

        try:
            cab_file = CabFile(arc_name, "w")

            for file_info in file_infos:
                rname = file_info.get_filename()
                a_name = file_info.get_filename_in_archive()
                with contextlib.suppress(IOError, OSError):
                    cab_file.write(rname, arcname=a_name)

            cab_file.close()
            return True

        except (IOError, OSError) as e:
            logger.debug("Archive creation IO error for %s: %s", arc_name, e)
        except Exception as e:
            logger.debug("Archive creation error for %s: %s", arc_name, e)

        return False



# -*- coding:utf-8 -*-
# Author: Kei Choi(hanul93@gmail.com)


# -------------------------------------------------------------------------
# Actual import modules
# -------------------------------------------------------------------------
import os
import sys

# Ensure project root is in sys.path for development mode
# This allows "from kicomav.plugins import ..." to work when running from source
_k2_dir = os.path.dirname(os.path.abspath(__file__))
_project_root = os.path.dirname(_k2_dir)
if _project_root not in sys.path:
    sys.path.insert(0, _project_root)

import contextlib
import glob
import hashlib
import time
import struct
import datetime
import gzip
import re
import tempfile
from optparse import OptionParser
from rich.console import Console
from rich.text import Text
import requests

# Configuration is automatically loaded by kavcore.config module
# when it is imported. The .env file is loaded from ~/.kicomav/.env

# Support both installed package and development mode
from kicomav import __version__ as KICOMAV_VERSION
from kicomav import __last_update__ as KICOMAV_BUILDDATE
from kicomav import kavcore
from kicomav.kavcore import k2engine as kavcore_k2engine
from kicomav.kavcore import k2const as kavcore_k2const
from kicomav.kavcore import k2security
from kicomav.kavcore import updater as kavcore_updater

# Alias for compatibility
kavcore.k2engine = kavcore_k2engine
kavcore.k2const = kavcore_k2const
kavcore.updater = kavcore_updater

# -------------------------------------------------------------------------
# Main constants
# -------------------------------------------------------------------------
KAV_VERSION = KICOMAV_VERSION
_parts = KICOMAV_BUILDDATE.split()
KAV_BUILDDATE = f"{_parts[1]} {int(_parts[2])} {_parts[4]}"
KAV_LASTYEAR = _parts[4]

g_options = None  # Options
g_delta_time = None  # Scan time
display_scan_result = {"Prev": {}, "Next": {}}  # Structure to prevent duplicate output
display_update_result = ""  # Structure to display compression results

PLUGIN_ERROR = False  # Variable used to format output nicely when plugin engine loading fails

# -------------------------------------------------------------------------
# Classes and functions for colored output on console (using rich)
# -------------------------------------------------------------------------
console = Console()

# Color style constants for rich
STYLE_RED = "bold red"
STYLE_GREEN = "green"
STYLE_CYAN = "bright_cyan"
STYLE_BRIGHT_BLUE = "bright_blue"
STYLE_BRIGHT_GREEN = "bright_green"
STYLE_GREY = "white"
STYLE_GREY_BOLD = "bold white"


def cprint(msg, style):
    """Print colored message using rich console."""
    console.print(msg, style=style, end="", highlight=False)


def print_error(msg):
    console.print(f"[{STYLE_RED}]Error:[/{STYLE_RED}] {msg}", highlight=False)


# -------------------------------------------------------------------------
# getch()
# Receives a single character input. Methods differ by operating system.
# -------------------------------------------------------------------------
def getch():
    if os.name == "nt":
        import msvcrt

        return msvcrt.getch()
    else:
        import tty
        import termios

        fd = sys.stdin.fileno()
        old_settings = termios.tcgetattr(fd)

        try:
            tty.setraw(sys.stdin.fileno())
            ch = sys.stdin.read(1)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)

        return ch


# ---------------------------------------------------------------------
# CreateFolder: Creates a file (generates the full path if it does not exist)
# ---------------------------------------------------------------------
def create_folder(path):
    if os.path.exists(path) is not False:
        return False

    t_dir = path

    # Proceed until a folder exists
    while 1:
        if os.path.exists(t_dir) is False:
            t_dir, tmp = os.path.split(t_dir)
        else:
            break

    makedir = path.replace(t_dir, "")
    mkdir_list = makedir.split(os.sep)

    for m in mkdir_list:
        if len(m) != 0:
            t_dir += os.sep + m
            # print dir # Create folder
            os.mkdir(t_dir)

    return True


# -------------------------------------------------------------------------
# log_print(msg)
# Writes a given message to a log file.
# -------------------------------------------------------------------------
def log_print(msg, file_mode="at"):
    global g_options

    if g_options != "NONE_OPTION":
        log_mode = False
        log_fname = "k2.log"  # Default log file name

        if g_options.log_filename:
            log_fname = g_options.log_filename
            log_mode = True

        if g_options.opt_log:
            log_mode = True

        if log_mode:
            with open(log_fname, file_mode) as fp:
                fp.write(msg)


# -------------------------------------------------------------------------
# print_k2logo()
# Displays the antivirus logo.
# -------------------------------------------------------------------------
def print_k2logo():
    logo = f"""KICOM Anti-Virus II (for {sys.platform.upper()}) Ver {KAV_VERSION} ({KAV_BUILDDATE})
Copyright (C) 1995-{KAV_LASTYEAR} Kei Choi. All rights reserved.
"""

    print("------------------------------------------------------------")
    cprint(logo, STYLE_CYAN)
    print("------------------------------------------------------------")


# -------------------------------------------------------------------------
# Redefines Python's option parser.
# Allows fine-tuned error messages.
# -------------------------------------------------------------------------
class OptionParsingError(RuntimeError):
    def __init__(self, msg):
        self.msg = msg


class OptionParsingExit(Exception):
    def __init__(self, status, msg):
        self.msg = msg
        self.status = status


class ModifiedOptionParser(OptionParser):
    def error(self, msg):
        raise OptionParsingError(msg)

    def exit(self, status=0, msg=None):
        raise OptionParsingExit(status, msg)


# -------------------------------------------------------------------------
# define_options()
# Defines the antivirus options.
# -------------------------------------------------------------------------
def define_options():
    usage = "Usage: %prog path[s] [options]"
    parser = ModifiedOptionParser(add_help_option=False, usage=usage)

    parser.add_option("-f", "--files", action="store_true", dest="opt_files", default=True)
    parser.add_option("-r", "--arc", action="store_true", dest="opt_arc", default=False)
    parser.add_option("-G", action="store_true", dest="opt_log", default=False)
    parser.add_option("", "--log", metavar="FILE", dest="log_filename")
    parser.add_option("-I", "--list", action="store_true", dest="opt_list", default=False)
    parser.add_option("-e", "--app", action="store_true", dest="opt_app", default=False)
    parser.add_option("-F", "--infp", metavar="PATH", dest="infp_path")

    # Assigns malware names during quarantine
    parser.add_option("", "--qname", action="store_true", dest="opt_qname", default=False)

    # Assigns Sha256 names during quarantine
    parser.add_option("", "--qhash", action="store_true", dest="opt_qhash", default=False)

    parser.add_option("-R", "--nor", action="store_true", dest="opt_nor", default=False)
    parser.add_option("-V", "--vlist", action="store_true", dest="opt_vlist", default=False)
    parser.add_option("-p", "--prompt", action="store_true", dest="opt_prompt", default=False)
    parser.add_option("-d", "--dis", action="store_true", dest="opt_dis", default=False)
    parser.add_option("-l", "--del", action="store_true", dest="opt_del", default=False)
    parser.add_option("", "--no-color", action="store_true", dest="opt_nocolor", default=False)
    parser.add_option("", "--move", action="store_true", dest="opt_move", default=False)
    parser.add_option("", "--copy", action="store_true", dest="opt_copy", default=False)
    parser.add_option("", "--update", action="store_true", dest="opt_update", default=False)
    parser.add_option("", "--verbose", action="store_true", dest="opt_verbose", default=False)
    parser.add_option("", "--sigtool", action="store_true", dest="opt_sigtool", default=False)
    parser.add_option("", "--debug", action="store_true", dest="opt_debug", default=False)
    parser.add_option("", "--password", metavar="PWD", dest="opt_password")
    parser.add_option("", "--parallel", action="store_true", dest="opt_parallel", default=False)
    parser.add_option("", "--workers", metavar="N", type="int", dest="opt_workers", default=0)
    parser.add_option("-?", "--help", action="store_true", dest="opt_help", default=False)

    return parser


# -------------------------------------------------------------------------
# parser_options()
# Parses the options for the antivirus
# -------------------------------------------------------------------------
def parser_options():
    parser = define_options()  # Define antivirus options

    if len(sys.argv) < 2:
        return "NONE_OPTION", None
    try:
        (options, args) = parser.parse_args()
        if len(args) == 0:
            return options, None
    except OptionParsingError as e:  # When invalid options are used
        return "ILLEGAL_OPTION", e.msg
    except OptionParsingExit as e:
        return "ILLEGAL_OPTION", e.msg

    return options, args


# -------------------------------------------------------------------------
# print_usage()
# Prints the usage of the antivirus
# -------------------------------------------------------------------------
def print_usage():
    print("\nUsage: k2.py path[s] [options]")


# -------------------------------------------------------------------------
# print_options()
# Prints the options of the antivirus
# -------------------------------------------------------------------------
def print_options():
    options_string = """Options:
        -f,  --files           scan files *
        -r,  --arc             scan archives
        -G,  --log=file        create log file
        -I,  --list            display all files
        -e,  --app             append to log file
        -F,  --infp=path       set infected quarantine folder
        -R,  --nor             do not recurse into folders
        -V,  --vlist           display virus list
        -p,  --prompt          prompt for action
        -d,  --dis             disinfect files
        -l,  --del             delete infected files
             --no-color        don't print with color
             --move            move infected files in quarantine folder
             --copy            copy infected files in quarantine folder
             --qname           quarantine by name of malware
             --qhash           quarantine by sha256 hash of malware
             --password=PWD    set password for encrypted archives
             --parallel        enable parallel file scanning
             --workers=N       number of worker threads (default: CPU count)
             --update          update
        -?,  --help            this help
                               * = default option"""

    print(options_string)


# -------------------------------------------------------------------------
# Get the installed version of a package (delegates to kavcore.updater)
# -------------------------------------------------------------------------
def get_installed_version(package_name):
    return kavcore_updater.get_installed_version(package_name)


# -------------------------------------------------------------------------
# get_latest_version()
# Get the latest version of a package from PyPI (delegates to kavcore.updater)
# -------------------------------------------------------------------------
def get_latest_version(package_name):
    return kavcore_updater.get_latest_version(package_name)


# -------------------------------------------------------------------------
# check_kicomav_update()
# Check if kicomav package needs update
# Returns: True if up-to-date, False if update needed
# -------------------------------------------------------------------------
def check_kicomav_update():
    installed_version = get_installed_version("kicomav")
    latest_version = get_latest_version("kicomav")

    if latest_version is None:
        cprint("[", STYLE_GREY)
        cprint("Warning", STYLE_BRIGHT_BLUE)
        cprint("] Cannot check latest version of ", STYLE_GREY)
        cprint("kicomav", STYLE_GREEN)
        cprint(" from PyPI\n\n", STYLE_GREY)
        return True  # Continue with signature update if PyPI check fails

    if installed_version != latest_version:
        cprint("[", STYLE_GREY)
        cprint("notice", STYLE_BRIGHT_BLUE)
        cprint(f"] A new release of ", STYLE_GREY)
        cprint("kicomav", STYLE_GREEN)
        cprint(" is available: ", STYLE_GREY)
        cprint(f"{installed_version}", STYLE_RED)
        cprint(" -> ", STYLE_GREY)
        cprint(f"{latest_version}\n", STYLE_GREEN)

        cprint("[", STYLE_GREY)
        cprint("notice", STYLE_BRIGHT_BLUE)
        cprint(f"] To update, run: ", STYLE_GREY)
        cprint("pip install --upgrade kicomav\n\n", STYLE_GREEN)
        return False  # Update required

    return True


# -------------------------------------------------------------------------
# get_signature_download_list()
# Get the list of signature files to update from SYSTEM_RULES_BASE path
# Returns: (down_list, cfg_files) - files to download and all files in cfg
# -------------------------------------------------------------------------
def get_signature_download_list(url, rules_path):
    down_list = []
    cfg_files = set()  # All files listed in update.cfg

    if not rules_path:
        return down_list, cfg_files

    with contextlib.suppress(Exception):
        # Download the update configuration file to a temp location
        temp_cfg_path = os.path.join(rules_path, "update.cfg")
        download_file(url, "update.cfg", rules_path)

        buf = open(temp_cfg_path, "r").read()
        # Format: [sha1] [filepath]
        p_lists = re.compile(r"([A-Fa-f0-9]{40}) (.+)")
        lines = p_lists.findall(buf)

        for line in lines:
            fhash = line[0]
            fname = line[1].strip()

            # Add to cfg_files set (for orphan detection)
            cfg_files.add(fname)

            # Normalize path separators for the current OS
            fname_normalized = fname.replace("/", os.sep).replace("\\", os.sep)

            # Special handling for whitelist.txt
            local_file = os.path.join(rules_path, fname_normalized)
            if fname == "whitelist.txt":
                if os.path.exists(local_file):
                    # Local whitelist exists - skip update (user settings priority)
                    continue
                else:
                    # Local whitelist doesn't exist - download it
                    down_list.append(fname)
                    continue

            # Compare the hash in the update config file with the local hash
            if chek_need_update(local_file, fhash) == 1:
                # If different, add to the update list
                down_list.append(fname)

        # Clean up update.cfg
        with contextlib.suppress(k2security.SecurityError):
            k2security.safe_remove_file(temp_cfg_path, rules_path)

    return down_list, cfg_files


# -------------------------------------------------------------------------
# get_local_files()
# Get all local files in rules_path (for orphan detection)
# -------------------------------------------------------------------------
def get_local_files(rules_path):
    local_files = set()
    for filepath in glob.glob(os.path.join(rules_path, "**", "*"), recursive=True):
        # Skip directories
        if os.path.isdir(filepath):
            continue
        # Convert to relative path from rules_path
        rel_path = os.path.relpath(filepath, rules_path)
        # Normalize to forward slashes for comparison
        rel_path = rel_path.replace("\\", "/")
        local_files.add(rel_path)
    return local_files


# -------------------------------------------------------------------------
# remove_orphan_files()
# Remove local files that are not in update.cfg (except whitelist.txt)
# -------------------------------------------------------------------------
def remove_orphan_files(rules_path, cfg_files):
    local_files = get_local_files(rules_path)

    # Normalize cfg_files paths
    cfg_files_normalized = {f.replace("\\", "/") for f in cfg_files}

    # Find orphan files (local only, not in cfg)
    orphan_files = local_files - cfg_files_normalized

    if not orphan_files:
        return

    for orphan in orphan_files:
        # Never delete whitelist.txt
        if orphan == "whitelist.txt":
            continue

        orphan_path = os.path.join(rules_path, orphan)
        try:
            # Determine base path for safe deletion
            if "/" in orphan:
                subdir = orphan.rsplit("/", 1)[0]
                base_path = os.path.join(rules_path, subdir)
            else:
                base_path = rules_path

            k2security.safe_remove_file(orphan_path, base_path)
            cprint(f"{orphan} ", STYLE_GREY)
            cprint("removed\n", STYLE_RED)
        except k2security.SecurityError:
            pass  # Skip files that can't be safely deleted


# -------------------------------------------------------------------------
# remove_empty_dirs()
# Remove empty directories in rules_path after orphan file cleanup
# -------------------------------------------------------------------------
def remove_empty_dirs(rules_path):
    # Walk bottom-up to remove nested empty directories
    for root, dirs, files in os.walk(rules_path, topdown=False):
        for dir_name in dirs:
            dir_path = os.path.join(root, dir_name)
            try:
                # Check if directory is empty
                if not os.listdir(dir_path):
                    os.rmdir(dir_path)
                    rel_path = os.path.relpath(dir_path, rules_path)
                    cprint(f"{rel_path}/ ", STYLE_GREY)
                    cprint("removed (empty)\n", STYLE_RED)
            except OSError:
                pass  # Skip directories that can't be removed


# -------------------------------------------------------------------------
# update_signatures()
# Update signature files (YARA rules) - CLI wrapper for kavcore.updater
# -------------------------------------------------------------------------
def update_signatures():
    cprint("", STYLE_GREY)

    # CLI progress callback
    def cli_progress(filename, status):
        if status == "downloading":
            cprint(f"{filename} ", STYLE_GREY)
        elif status == "updated":
            cprint(" update\n", STYLE_GREEN)
        elif status == "removed":
            cprint(f"{filename} ", STYLE_GREY)
            cprint("removed\n", STYLE_RED)

    # Use kavcore.updater for the actual update
    result = kavcore_updater.update_signatures(progress_callback=cli_progress)

    if not result.success:
        for error in result.errors:
            cprint("[", STYLE_GREY)
            cprint("Warning", STYLE_BRIGHT_BLUE)
            cprint(f"] {error}\n", STYLE_GREY)
        return

    if not result.updated_files and not result.removed_files:
        cprint("[", STYLE_GREY)
        cprint("No updates available", STYLE_GREEN)
        cprint("]\n", STYLE_GREY)
        return

    cprint("\n[", STYLE_GREY)
    cprint("Signature updates complete", STYLE_GREEN)
    cprint("]\n", STYLE_GREY)


# -------------------------------------------------------------------------
# update_kicomav()
# Updates Kicom Antivirus to the latest version
# Two-phase update: (1) Package update check, (2) Signature update
# -------------------------------------------------------------------------
def update_kicomav(path):
    print()

    try:
        # (1) Check kicomav package version
        if not check_kicomav_update():
            # Package update required - exit after showing pip command
            return

        # (2) Update signatures (YARA rules)
        update_signatures()

    except KeyboardInterrupt:
        cprint("\n[", STYLE_GREY)
        cprint("Update Stop", STYLE_GREY_BOLD)
        cprint("]\n", STYLE_GREY)
    except Exception as e:
        if g_options.opt_verbose:
            print(e)
        cprint("\n[", STYLE_GREY)
        cprint("Update failed", STYLE_RED)
        cprint("]\n", STYLE_GREY)


# Display update progress
def hook(blocknumber, blocksize, totalsize):
    cprint(".", STYLE_GREY)


# Downloads a single file (CWE-434, CWE-918 prevention)
# Supports subdirectory paths like "yara/subdir/file.yar"
def download_file(url, filename, path, gz=False, fnhook=None):
    # Normalize path separators to forward slash
    filename = filename.replace("\\", "/")

    # Split into directory path and actual filename
    if "/" in filename:
        subdir, actual_filename = filename.rsplit("/", 1)
        # Validate for path traversal
        if ".." in subdir:
            cprint(f" {filename} - invalid path: path traversal detected\n", STYLE_RED)
            return
    else:
        subdir = ""
        actual_filename = filename

    # CWE-434: Validate actual filename (without path) before download
    try:
        k2security.validate_download_filename(actual_filename)
    except k2security.SecurityError as e:
        cprint(f" {filename} - invalid filename: {e}\n", STYLE_RED)
        return

    # Create subdirectory if needed
    if subdir:
        subdir_path = os.path.join(path, subdir)
        os.makedirs(subdir_path, exist_ok=True)

    rurl = url

    # Convert the list in the update config file to a URL address
    rurl += filename.replace("\\", "/")
    if gz:
        rurl += ".gz"

    # Get the full path of the file to save
    pwd = os.path.join(path, filename)

    if gz:
        pwd += ".gz"

    if fnhook is not None:
        cprint(f"{filename} ", STYLE_GREY)

    # Download the file (CWE-918: SSRF prevention with URL validation)
    try:
        # Use the subdirectory path as base for security validation
        base_path = os.path.join(path, subdir) if subdir else path
        k2security.safe_download_file(rurl, pwd, base_path, fnhook=fnhook)
    except k2security.SecurityError as e:
        cprint(f" download failed: {e}\n", STYLE_RED)
        return

    if gz:
        data = gzip.open(pwd, "rb").read()
        fname = os.path.join(path, filename)
        open(fname, "wb").write(data)
        # Delete the gz file (CWE-73 safe deletion)
        base_path = os.path.join(path, subdir) if subdir else path
        with contextlib.suppress(k2security.SecurityError):
            k2security.safe_remove_file(pwd, base_path)

    if fnhook is not None:
        cprint(" update\n", STYLE_GREEN)


# Downloads k2.exe (CWE-434, CWE-918 prevention)
def download_file_k2(url, filename, path, gz=False, fnhook=None):
    # CWE-434: Validate filename before download
    try:
        k2security.validate_download_filename(filename)
    except k2security.SecurityError as e:
        cprint(f" {filename} - invalid filename: {e}\n", STYLE_RED)
        return None

    rurl = url

    # Convert the list in the update config file to a URL address
    rurl += filename.replace("\\", "/")
    if gz:
        rurl += ".gz"

    # Get the full path of the file to save
    pwd = os.path.join(path, filename)
    if gz:
        pwd += ".gz"

    if fnhook is not None:
        cprint(f"{filename} ", STYLE_GREY)

    # Download the file (CWE-918: SSRF prevention with URL validation)
    try:
        k2security.safe_download_file(rurl, pwd, path, fnhook=fnhook)
    except k2security.SecurityError as e:
        cprint(f" download failed: {e}\n", STYLE_RED)
        return None

    if gz:
        data = gzip.open(pwd, "rb").read()
        # Use mkstemp instead of mktemp to prevent race condition (CWE-377)
        fd, fname = tempfile.mkstemp(prefix="ktmp", suffix=".exe")
        try:
            os.write(fd, data)
        finally:
            os.close(fd)
        # Delete the gz file (CWE-73 safe deletion)
        with contextlib.suppress(k2security.SecurityError):
            k2security.safe_remove_file(pwd, path)

    if fnhook is not None:
        cprint(" update\n", STYLE_GREEN)

    return fname


# Compare the hash in the update config file with the local hash
def chek_need_update(file, hash):
    with contextlib.suppress(IOError):
        with open(file, "rb") as fp:
            data = fp.read()
        # Compare the hash (case-insensitive comparison)
        s = hashlib.sha1()
        s.update(data)
        if s.hexdigest().lower() == hash.lower():
            return 0  # Not an update target

    return 1  # Update target


# -------------------------------------------------------------------------
# Callback function for listvirus
# -------------------------------------------------------------------------
def listvirus_callback(plugin_name, vnames):
    terminal_width = get_terminal_sizex()
    engine_str = f"[{plugin_name}]"
    engine_len = len(engine_str)

    # Available space for malware name (1 space before engine name)
    available_width = terminal_width - engine_len - 1

    for vname in vnames:
        if len(vname) <= available_width:
            # Malware name fits, right-align engine name
            padding = terminal_width - len(vname) - engine_len
            print(f"{vname}{' ' * padding}{engine_str}")
        else:
            # Truncate malware name with ' ... ' in the middle
            ellipsis = " ... "
            half_len = (available_width - len(ellipsis)) // 2
            first_part = vname[:half_len]
            last_part = vname[-(available_width - half_len - len(ellipsis)) :]
            truncated = f"{first_part}{ellipsis}{last_part}"
            print(f"{truncated} {engine_str}")


# -------------------------------------------------------------------------
# Function to display malware results in a single line
# -------------------------------------------------------------------------
def get_terminal_sizex():
    default_sizex = 80

    # Source: https://gist.github.com/jtriley/1108174
    if os.name == "nt":
        with contextlib.suppress(Exception):
            from ctypes import windll, create_string_buffer

            h = windll.kernel32.GetStdHandle(-12)
            csbi = create_string_buffer(22)
            res = windll.kernel32.GetConsoleScreenBufferInfo(h, csbi)
            if res:
                (
                    bufx,
                    bufy,
                    curx,
                    cury,
                    wattr,
                    left,
                    top,
                    right,
                    bottom,
                    maxx,
                    maxy,
                ) = struct.unpack("hhhhHhhhhhh", csbi.raw)
                sizex = right - left + 1
                # sizey = bottom - top + 1
                return sizex
    else:

        def ioctl_GWINSZ(fd):
            with contextlib.suppress(Exception):
                import fcntl
                import termios

                cr = struct.unpack("hh", fcntl.ioctl(fd, termios.TIOCGWINSZ, "1234"))
                return cr

        cr = ioctl_GWINSZ(0) or ioctl_GWINSZ(1) or ioctl_GWINSZ(2)
        if not cr:
            with contextlib.suppress(Exception):
                fd = os.open(os.ctermid(), os.O_RDONLY)
                cr = ioctl_GWINSZ(fd)
                os.close(fd)

        if not cr:
            try:
                cr = (os.environ["LINES"], os.environ["COLUMNS"])
            except Exception:
                return default_sizex
        return int(cr[1])  # , int(cr[0])

    return default_sizex  # default


def convert_display_filename(real_filename):
    # Name for display
    # Use stdout encoding for both encode and decode to avoid mismatch
    stdout_encoding = sys.stdout.encoding or sys.getdefaultencoding()
    display_filename = real_filename.encode(stdout_encoding, "replace")

    # In Python 3, bytes[0] returns int, so compare with ord() values
    if display_filename and display_filename[0] in [ord("/"), ord("\\")]:
        return display_filename[1:].decode(stdout_encoding, "replace")
    else:
        return display_filename.decode(stdout_encoding, "replace")


def get_display_width(s):
    """Calculate display width of a string (wide characters count as 2)."""
    import unicodedata

    width = 0
    for char in s:
        ea = unicodedata.east_asian_width(char)
        if ea in ("F", "W"):  # Fullwidth or Wide
            width += 2
        else:
            width += 1
    return width


def truncate_to_width(s, target_width):
    """Truncate string to fit within target display width."""
    import unicodedata

    width = 0
    result = []
    for char in s:
        ea = unicodedata.east_asian_width(char)
        char_width = 2 if ea in ("F", "W") else 1
        if width + char_width > target_width:
            break
        result.append(char)
        width += char_width
    return "".join(result)


def get_suffix_by_width(s, target_width):
    """Get suffix of string that fits within target display width."""
    import unicodedata

    chars = list(s)
    width = 0
    start_idx = len(chars)
    for i in range(len(chars) - 1, -1, -1):
        ea = unicodedata.east_asian_width(chars[i])
        char_width = 2 if ea in ("F", "W") else 1
        if width + char_width > target_width:
            break
        width += char_width
        start_idx = i
    return "".join(chars[start_idx:])


def display_line(filename, message, message_color):
    max_sizex = get_terminal_sizex() - 1
    filename += " "
    filename = convert_display_filename(filename)
    disp_width = get_display_width(filename)
    len_msg = len(message)

    if disp_width + 1 + len_msg < max_sizex:
        fname = f"{filename}"
    else:
        fname = truncate_filename_with_ellipsis(max_sizex, len_msg, filename, disp_width)

    # Use Text object to combine multiple styles in one line
    text = Text()
    text.append(f"{fname} ", style=STYLE_GREY)
    text.append(message, style=message_color)
    console.print(text)


def truncate_filename_with_ellipsis(max_sizex, len_msg, filename, disp_width):
    """Function to shorten long filenames to fit the screen size and insert ... in the middle"""
    able_size = max_sizex - len_msg
    able_size -= 5  # " ... "
    if able_size <= 0:
        return filename[:10] + " ... "
    half_size = able_size // 2
    fname1 = truncate_to_width(filename, half_size)
    fname2 = get_suffix_by_width(filename, able_size - half_size)

    return f"{fname1} ... {fname2}"


# -------------------------------------------------------------------------
# Callback function for scan
# -------------------------------------------------------------------------
def scan_callback(ret_value):
    global g_options
    global display_scan_result  # Structure to temporarily hold output

    from kicomav.plugins import kernel

    fs = ret_value["file_struct"]

    if len(fs.get_additional_filename()) != 0:
        f2 = convert_display_filename(fs.get_additional_filename())
        disp_name = f"{fs.get_master_filename()} ({f2})"
    else:
        disp_name = f"{fs.get_master_filename()}"

    if ret_value["result"]:
        if ret_value["scan_state"] == kernel.INFECTED:
            state = "infected"
            message_color = STYLE_RED
        elif ret_value["scan_state"] == kernel.SUSPECT:
            state = "suspect"
            message_color = STYLE_RED
        elif ret_value["scan_state"] == kernel.WARNING:
            state = "warning"
            message_color = STYLE_RED
        elif ret_value["scan_state"] == kernel.IDENTIFIED:
            state = "identified"
            message_color = STYLE_GREEN
        else:
            state = "unknown"
            message_color = STYLE_RED

        vname = ret_value["virus_name"]
        message = f"{state} : {vname}"
    elif ret_value["scan_state"] == kernel.ERROR:
        message = ret_value["virus_name"]
        message_color = STYLE_CYAN
    else:
        message = "ok"
        message_color = STYLE_GREY_BOLD

    # In normal cases, there is a possibility of duplication due to /<...> paths
    # Adjusted to prevent duplicate output
    if message == "ok":
        d_prev = display_scan_result.get("Prev", {})
        if d_prev == {}:
            update_scan_result_cache(disp_name, "Prev", message, message_color)
        elif d_prev["disp_name"] != disp_name:
            d_next = display_scan_result.get("Next", {})
            if d_next == {}:
                update_scan_result_cache(disp_name, "Next", message, message_color)
            elif d_next["disp_name"] != disp_name:
                # Print Prev
                print_and_clear_cached_result(d_prev, display_scan_result, "Prev")
                log_print(f"{d_prev['disp_name']}	{d_prev['message']}\n")

                # Move Next to Prev
                update_scan_result_cache(d_next["disp_name"], "Prev", d_next["message"], d_next["message_color"])

                update_scan_result_cache(disp_name, "Next", message, message_color)
    else:  # If malware is found, print all stored outputs
        print_display_scan_result(disp_name, message, message_color)

    if g_options.opt_move is False and g_options.opt_prompt:  # Is the prompt option set?
        while ret_value["result"]:
            if ret_value["scan_state"] == kernel.INFECTED:
                msg = "Disinfect/Delete/Ignore/Quit? (d/l/i/q) : "
            else:
                msg = "Delete/Ignore/Quit? (l/i/q) : "

            cprint(msg, STYLE_CYAN)
            log_print(msg)

            ch = getch().lower()
            print(ch)
            log_print(ch + "\n")

            if ret_value["scan_state"] == kernel.INFECTED and ch == "d":
                return kavcore.k2const.K2_ACTION_DISINFECT
            elif ch == "l":
                return kavcore.k2const.K2_ACTION_DELETE
            elif ch == "i":
                return kavcore.k2const.K2_ACTION_IGNORE
            elif ch == "q":
                return kavcore.k2const.K2_ACTION_QUIT
    elif g_options.opt_dis:  # Disinfect option
        return kavcore.k2const.K2_ACTION_DISINFECT
    elif g_options.opt_del:  # Delete option
        return kavcore.k2const.K2_ACTION_DELETE

    return kavcore.k2const.K2_ACTION_IGNORE


def update_scan_result_cache(disp_name, result_dict, message, message_color):
    """Function to temporarily store scan results"""
    global display_scan_result  # Structure to temporarily hold output

    display_scan_result[result_dict]["disp_name"] = disp_name
    display_scan_result[result_dict]["message"] = message
    display_scan_result[result_dict]["message_color"] = message_color


# Prints the results of the display_scan_result structure
def print_display_scan_result(disp_name, message, message_color):
    global display_scan_result  # Structure to temporarily hold output

    # Print Prev
    d_prev = display_scan_result.get("Prev", {})
    if d_prev != {} and d_prev["disp_name"] != disp_name:
        print_and_clear_cached_result(d_prev, display_scan_result, "Prev")
    # Print Next
    d_next = display_scan_result.get("Next", {})
    if d_next != {} and d_next["disp_name"] != disp_name:
        print_and_clear_cached_result(d_next, display_scan_result, "Next")
    # Print the final result
    if disp_name:
        display_line(disp_name, message, message_color)
        log_print(f"{disp_name}\t{message}\n")


def print_and_clear_cached_result(cached_result, display_scan_result, cache_key):
    """Function to print and clear cached scan results"""
    display_line(cached_result["disp_name"], cached_result["message"], cached_result["message_color"])
    log_print(f"{cached_result['disp_name']}\t{cached_result['message']}\n")
    display_scan_result[cache_key] = {}


# -------------------------------------------------------------------------
# Callback function for disinfect
# -------------------------------------------------------------------------
def disinfect_callback(ret_value, action_type):
    fs = ret_value["file_struct"]
    message = ""

    if len(fs.get_additional_filename()) != 0:
        disp_name = f"{fs.get_master_filename()} ({fs.get_additional_filename()})"
    else:
        disp_name = f"{fs.get_master_filename()}"

    if fs.is_modify():  # Modification successful?
        if action_type == kavcore.k2const.K2_ACTION_DISINFECT:
            message = "disinfected"
        elif action_type == kavcore.k2const.K2_ACTION_DELETE:
            message = "deleted"

        message_color = STYLE_GREEN
    else:
        if action_type == kavcore.k2const.K2_ACTION_DISINFECT:
            message = "disinfection failed"
        elif action_type == kavcore.k2const.K2_ACTION_DELETE:
            message = "deletion failed"

        message_color = STYLE_RED

    display_line(disp_name, message, message_color)
    log_print(f"{disp_name}\t{message}\n")


# -------------------------------------------------------------------------
# Callback function for update
# -------------------------------------------------------------------------
def update_callback(ret_file_info, is_success):
    global display_update_result

    # Print results that were not output
    print_display_scan_result(None, None, None)

    if ret_file_info.is_modify():  # If modified, print results
        if len(ret_file_info.get_additional_filename()) != 0:
            disp_name = f"{ret_file_info.get_master_filename()} ({ret_file_info.get_additional_filename()})"
        else:
            disp_name = f"{ret_file_info.get_master_filename()}"

        if is_success:
            if os.path.exists(ret_file_info.get_filename()):
                message = "updated"
            else:
                message = "deleted"

            message_color = STYLE_GREEN
        else:
            message = "update failed"
            message_color = STYLE_RED

        if display_update_result != disp_name:  # Do not print if the same as previous output
            display_line(disp_name, message, message_color)
            log_print(f"{disp_name}\t{message}\n")

            display_update_result = disp_name


# -------------------------------------------------------------------------
# Callback function for quarantine
# -------------------------------------------------------------------------
def quarantine_callback(filename, is_success, q_type):
    from kicomav.plugins import kernel

    q_message = {
        kavcore.k2const.K2_QUARANTINE_MOVE: ["quarantined", "quarantine failed"],
        kavcore.k2const.K2_QUARANTINE_COPY: ["copied", "copy failed"],
    }

    msg = q_message[q_type]

    disp_name = filename

    if is_success:
        message = msg[0]  # Success
        message_color = STYLE_GREEN
    else:
        message = msg[1]  # Failure
        message_color = STYLE_RED

    display_line(disp_name, message, message_color)
    log_print(f"{disp_name}\t{message}\n")


# -------------------------------------------------------------------------
# Callback function for plugin engine loading failure
# -------------------------------------------------------------------------
def import_error_callback(module_name):
    global PLUGIN_ERROR
    global g_options

    if g_options.opt_debug and not PLUGIN_ERROR:
        PLUGIN_ERROR = True
        print()
        print_error(f"Invalid plugin: '{module_name}'")


# -------------------------------------------------------------------------
# print_result(result)
# Prints the results of the malware scan.
# Input: result - Malware scan results
# -------------------------------------------------------------------------
def print_result(result):
    global g_options
    global g_delta_time

    print()
    print()

    cprint("Results:\n", STYLE_GREY_BOLD)
    cprint(f"Folders           :{result['Folders']}\n", STYLE_GREY_BOLD)
    cprint(f"Files             :{result['Files']}\n", STYLE_GREY_BOLD)
    cprint(f"Packed            :{result['Packed']}\n", STYLE_GREY_BOLD)
    cprint(f"Infected files    :{result['Infected_files']}\n", STYLE_GREY_BOLD)
    cprint(f"Suspect files     :{result['Suspect_files']}\n", STYLE_GREY_BOLD)
    cprint(f"Warnings          :{result['Warnings']}\n", STYLE_GREY_BOLD)
    cprint(f"Identified viruses:{result['Identified_viruses']}\n", STYLE_GREY_BOLD)
    if result["Disinfected_files"]:
        cprint(f"Disinfected files :{result['Disinfected_files']}\n", STYLE_GREY_BOLD)
    elif result["Deleted_files"]:
        cprint(f"Deleted files     :{result['Deleted_files']}\n", STYLE_GREY_BOLD)
    cprint(f"I/O errors        :{result['IO_errors']}\n", STYLE_GREY_BOLD)

    # Display scan time
    t = str(g_delta_time).split(":")
    t_h = int(float(t[0]))
    t_m = int(float(t[1]))
    t_s = int(float(t[2]))
    cprint(f"Scan time         :{t_h:02d}:{t_m:02d}:{t_s:02d}\n", STYLE_GREY_BOLD)

    print()


# -------------------------------------------------------------------------
# main()
# -------------------------------------------------------------------------
def main():
    global console
    global g_options

    # Parse options
    options, args = parser_options()
    g_options = options  # Set global options

    # Handle --no-color option
    if not isinstance(options, str) and options.opt_nocolor:
        console = Console(no_color=True)

    # Display logo
    print_k2logo()

    # Invalid options?
    if options == "NONE_OPTION":  # No options provided
        return print_usage_and_options()
    elif options == "ILLEGAL_OPTION":  # Undefined options used
        print_usage()
        print(f"Error: {args}")
        return 0

    # Program's running folder (use __file__ to get actual module location)
    k2_pwd = os.path.dirname(os.path.abspath(__file__))

    # Help option or no arguments
    if options.opt_help or not args:
        # Update scenario with no arguments?
        if options.opt_update:
            update_kicomav(k2_pwd)
            return 0

        if not options.opt_vlist:  # No malware list output
            return print_usage_and_options()

    # Create log file
    if g_options.opt_app is False:
        log_print("#\n# KicomAV scan report\n#\n", "wt")  # Create new file if not append mode
    else:
        log_print("\n#\n# KicomAV scan report\n#\n")

    log_print(f"# Time: {time.ctime()}\n")

    log_print("# Command line: ")
    for argv in sys.argv[1:]:
        log_print(f"{argv} ")
    log_print("\n")

    logo = f"KICOM Anti-Virus II (for {sys.platform.upper()}) Ver {KAV_VERSION} ({KAV_BUILDDATE})"
    log_print(f"# {logo}\n")

    # Post-update scanning
    if options.opt_update:
        update_kicomav()

    # Set quarantine folder
    if options.infp_path:
        path = os.path.abspath(options.infp_path)
        path = os.path.normcase(path)
        create_folder(path)
        options.infp_path = path

    with console.status("Loading KicomAV...", spinner="dots"):
        # Start antivirus engine
        k2 = kavcore.k2engine.Engine()  # Engine class

        # Set plugin engine
        plugins_path = os.path.join(k2_pwd, "plugins")
        if not k2.set_plugins(plugins_path, import_error_callback):
            return print_engine_error("KICOM Anti-Virus Engine set_plugins")
        kav = k2.create_instance()  # Create antivirus engine instance
        if not kav:
            return print_engine_error("KICOM Anti-Virus Engine create_instance")
        kav.set_options(options)  # Set options

        if not kav.init(import_error_callback):  # Initialize all plugin engines
            return print_engine_error("KICOM Anti-Virus Engine init")

    if options.opt_debug and PLUGIN_ERROR:
        print()

    # Print engine version
    c = kav.get_version()
    msg = f"\rLast updated {c.ctime()} UTC\n"
    cprint(msg, STYLE_GREY)

    # Print the number of diagnosable/treatable malware
    num_sig = format(kav.get_signum(), ",")
    msg = f"Signature number: {num_sig}\n\n"
    cprint(msg, STYLE_GREY)

    log_print(f"# Signature number: {num_sig}\n")
    log_print("#\n\n")

    if options.opt_vlist is True:  # Display malware list?
        kav.listvirus(listvirus_callback)
    elif args:
        scan_paths_and_print_result(kav, args, k2)
    kav.uninit()


def scan_paths_and_print_result(kav, args, k2_engine=None):
    """Scan the given paths and print the results"""
    global g_options

    kav.set_result()  # Initialize malware scan results

    # Check scan start time
    start_time = datetime.datetime.now()

    # Determine if parallel scanning should be used
    use_parallel = g_options.opt_parallel
    max_workers = g_options.opt_workers if g_options.opt_workers > 0 else (os.cpu_count() or 4)

    interrupted = False

    # Scan paths (supports multiple paths)
    try:
        for scan_path in args:  # First argument excluding options is the target
            scan_path = os.path.abspath(scan_path)

            if os.path.exists(scan_path):  # Does the folder or file exist?
                if use_parallel and k2_engine:
                    # Parallel scanning mode
                    status_msg = f"Parallel scanning ({max_workers} workers)..."
                    with console.status(status_msg, spinner="dots"):
                        ret = kav.scan_parallel(
                            scan_path,
                            max_workers,
                            k2_engine,
                            scan_callback,
                            disinfect_callback,
                            update_callback,
                            quarantine_callback,
                        )
                        if ret == 1:  # Interrupted
                            interrupted = True
                            break
                else:
                    # Sequential scanning mode (default)
                    with console.status("Scanning...", spinner="dots"):
                        kav.scan(
                            scan_path,
                            scan_callback,
                            disinfect_callback,
                            update_callback,
                            quarantine_callback,
                        )
            else:
                # Print results not displayed
                print_display_scan_result(None, None, None)
                print_error(f"Invalid path: '{scan_path}'")
    except KeyboardInterrupt:
        interrupted = True

    # Check scan end time
    end_time = datetime.datetime.now()

    global g_delta_time
    g_delta_time = end_time - start_time

    # Print results (ignore Ctrl+C during output to ensure results are shown)
    import signal

    original_handler = signal.getsignal(signal.SIGINT)
    signal.signal(signal.SIGINT, signal.SIG_IGN)  # Ignore Ctrl+C
    try:
        print_display_scan_result(None, None, None)
        ret = kav.get_result()
        print_result(ret)
    finally:
        signal.signal(signal.SIGINT, original_handler)  # Restore handler


def print_usage_and_options():
    """Print usage and options"""
    print_usage()
    print_options()
    return 0


def print_engine_error(error_msg):
    """Print engine initialization error"""
    print()
    print_error(error_msg)
    return 0


if __name__ == "__main__":
    main()



# -*- coding:utf-8 -*-
# Author: Kei Choi(hanul93@gmail.com)


# -------------------------------------------------------------------------
# Constants for specifying virus removal actions
# Used as return value from scan callback function
# -------------------------------------------------------------------------
K2_ACTION_IGNORE = 0
K2_ACTION_DISINFECT = 1
K2_ACTION_DELETE = 2
K2_ACTION_QUIT = 3

# -------------------------------------------------------------------------
# Constants for specifying virus isolation status
# -------------------------------------------------------------------------
K2_QUARANTINE_MOVE = 0
K2_QUARANTINE_COPY = 1


# -------------------------------------------------------------------------
# Custom exception for plugin unexpected errors
# Used to propagate errors to k2engine for IO_errors counting
# -------------------------------------------------------------------------
class PluginUnexpectedError(Exception):
    """Exception raised when a plugin encounters an unexpected error.

    This exception is raised after logging the error, allowing k2engine
    to catch it and increment the IO_errors counter.
    """

    pass



# -*- coding:utf-8 -*-
# Author: Kei Choi(hanul93@gmail.com)

"""
Script Engine Plugin

This plugin handles script format detection, malware detection, and extraction.
"""

import contextlib
import hashlib
import logging
import os
import re

from kicomav.plugins import kavutil
from kicomav.plugins import kernel
from kicomav.kavcore import k2security
from kicomav.kavcore.plugin_base import ArchivePluginBase

# Module logger
logger = logging.getLogger(__name__)

KICOMAV_BAT_MAGIC = b"<KicomAV:BAT>"


# -------------------------------------------------------------------------
# KavMain class
# -------------------------------------------------------------------------
class KavMain(ArchivePluginBase):
    """Script malware detector and archive handler plugin.

    This plugin provides functionality for:
    - Detecting script format (BAT, JavaScript, IFrame)
    - Scanning for script-based malware
    - Extracting scripts for further analysis
    """

    def __init__(self):
        """Initialize the Script Engine plugin."""
        super().__init__(
            author="Kei Choi",
            version="1.0",
            title="Script Engine",
            kmd_name="script",
        )
        self.p_text_format = None
        self.p_script_head = None
        self.p_script_in_html = None
        self.p_http = None
        self.p_script_cmt1 = None
        self.p_script_cmt2 = None
        self.p_script_cmt3 = None
        self.p_space = None
        self.p_bat_cmt1 = None
        self.p_bat_cmt2 = None

    def _custom_init(self) -> int:
        """Custom initialization for Script plugin.

        Returns:
            0 for success
        """
        # Check if the file format is text
        self.p_text_format = re.compile(r"\s*@?(\w+)", re.IGNORECASE)

        # Check if the file starts with <script, <iframe
        self.p_script_head = re.compile(r"\s*<\s*(script|iframe)", re.IGNORECASE)

        # Check if the script/iframe information is inside the html
        s = r"<\s*(script|iframe).*?>([\d\D]*?)<\s*/(script|iframe)\s*>"
        self.p_script_in_html = re.compile(s, re.IGNORECASE)

        # Regular expression for removing comments and spaces
        self.p_http = re.compile(rb"https?://")
        self.p_script_cmt1 = re.compile(rb"//.*")
        self.p_script_cmt2 = re.compile(rb"/\*.*?\*/", re.DOTALL)
        self.p_script_cmt3 = re.compile(rb"(#|\bREM\b).*", re.IGNORECASE)
        self.p_space = re.compile(rb"[\s]")

        # BAT comments
        self.p_bat_cmt1 = re.compile(rb"\bREM\s+.*", re.IGNORECASE)
        self.p_bat_cmt2 = re.compile(rb"[\^\`]", re.IGNORECASE)

        return 0

    def getinfo(self):
        """Get plugin information.

        Returns:
            Dictionary containing plugin metadata
        """
        info = super().getinfo()
        s_num = 0
        if kavutil.handle_pattern_md5:
            s_num = kavutil.handle_pattern_md5.get_sig_num("script")
        info["sig_num"] = s_num
        return info

    def listvirus(self):
        """Get list of detectable viruses.

        Returns:
            List of virus names
        """
        vlist = kavutil.handle_pattern_md5.get_sig_vlist("script")
        if vlist is None:
            vlist = []
        vlist.sort()
        return vlist

    def format(self, filehandle, filename, filename_ex):
        """Analyze and detect script format.

        Args:
            filehandle: File data (memory mapped)
            filename: Path to file
            filename_ex: Extended filename info

        Returns:
            Dictionary with format info, or None if not recognized
        """
        try:
            fileformat = {}
            mm = filehandle

            buf = mm.read(4096) if hasattr(mm, "read") else mm[:4096]

            if kavutil.is_textfile(buf):
                buf_str = buf.decode("latin-1")
                if obj := self.p_text_format.match(buf_str):
                    t = obj.groups()[0].lower()
                    if t in ["cd", "echo"]:
                        return {"ff_bat": "BAT"}
                elif buf_str[:13] == KICOMAV_BAT_MAGIC.decode("latin-1"):
                    return {"ff_bat": "BAT"}

                if obj := self.p_script_head.match(buf_str):
                    if hasattr(mm, "seek"):
                        mm.seek(0)
                        mm_buf = mm.read()
                    else:
                        mm_buf = mm[:]
                    mm_buf = mm_buf.decode("latin-1")
                    if obj_script := self.p_script_in_html.search(mm_buf):
                        buf_script = obj_script.groups()[1].strip()
                        n_buf_script = len(buf_script)
                        fileformat["size"] = n_buf_script

                        if n_buf_script:
                            ret = (
                                {"ff_script": fileformat}
                                if obj_script.groups()[0].lower() == "script"
                                else {"ff_iframe": fileformat}
                            )
                        elif obj_script.groups()[0].lower() == "script":
                            ret = {"ff_script_external": fileformat}
                        else:
                            ret = {"ff_iframe_external": fileformat}
                    else:
                        fileformat["size"] = 0

                        if obj.group().lower().find("script") != -1:
                            ret = {"ff_script_external": fileformat}
                        else:
                            ret = {"ff_iframe_external": fileformat}

                    return ret

        except (IOError, OSError) as e:
            logger.debug("Format detection IO error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error in format detection for %s: %s", filename, e)

        return None

    def scan(self, filehandle, filename, fileformat, filename_ex):
        """Scan for malware.

        Args:
            filehandle: File data (memory mapped)
            filename: Path to file
            fileformat: Format info from format() method
            filename_ex: Extended filename info

        Returns:
            Tuple of (found, malware_name, malware_id, result)
        """
        try:
            mm = filehandle

            if not (
                "ff_html" in fileformat
                or "ff_script" in fileformat
                or "ff_iframe" in fileformat
                or "ff_bat" in fileformat
                or "ff_script_external" in fileformat
                or "ff_iframe_external" in fileformat
            ):
                return False, "", -1, kernel.NOT_FOUND

            if hasattr(mm, "seek"):
                mm.seek(0)
                mm_buf = mm.read()
            else:
                mm_buf = mm[:]

            if "ff_bat" in fileformat and mm_buf[:13] == KICOMAV_BAT_MAGIC:
                p = re.compile(rb"set\s+(\w+)=", re.IGNORECASE)
                t_set = p.findall(mm_buf)
                t_count = 0
                for k in t_set:
                    p = re.compile(b"echo\s+.+?%s" % k)
                    if p.search(mm_buf):
                        t_count += 1

                if t_count > 5:
                    return True, "Trojan.BAT.Agent.gen", 0, kernel.INFECTED

            if kavutil.is_textfile(mm_buf[:4096]):
                buf = mm_buf

                buf = self.p_http.sub(b"", buf)
                buf = self.p_script_cmt1.sub(b"", buf)

                pos1 = buf.find(b"/*")
                pos2 = buf.rfind(b"*/") if pos1 != -1 else -1
                if 0 <= pos1 < pos2:
                    buf = self.p_script_cmt2.sub(b"", buf)

                buf = self.p_script_cmt3.sub(b"", buf)
                buf = self.p_space.sub(b"", buf)
                buf = buf.lower()

                size = len(buf)
                if kavutil.handle_pattern_md5.match_size("script", size):
                    fmd5 = hashlib.md5(buf).hexdigest()
                    if vname := kavutil.handle_pattern_md5.scan("script", size, fmd5):
                        return True, vname, 0, kernel.INFECTED

        except (IOError, OSError) as e:
            logger.debug("Scan IO error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error scanning %s: %s", filename, e)

        return False, "", -1, kernel.NOT_FOUND

    def disinfect(self, filename, malware_id):
        """Disinfect malware.

        Args:
            filename: Path to infected file
            malware_id: Malware ID to disinfect

        Returns:
            True if successful, False otherwise
        """
        try:
            if malware_id == 0:
                filename_dir = os.path.dirname(filename) or os.getcwd()
                k2security.safe_remove_file(filename, filename_dir)
                return True

        except (IOError, OSError, k2security.SecurityError) as e:
            logger.debug("Disinfect error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error disinfecting %s: %s", filename, e)

        return False

    def arclist(self, filename, fileformat, password=None):
        """List files in the archive (scripts in document).

        Args:
            filename: Path to script file
            fileformat: Format info from format() method

        Returns:
            List of [engine_id, filename] pairs
        """
        file_scan_list = []

        if "ff_script" in fileformat:
            file_scan_list.append(["arc_script", "JavaScript"])
        elif "ff_iframe" in fileformat:
            file_scan_list.append(["arc_iframe", "IFrame"])
        elif "ff_bat" in fileformat:
            file_scan_list.append(["arc_bat", "BAT"])

        return file_scan_list

    def unarc(self, arc_engine_id, arc_name, fname_in_arc):
        """Extract a file from the archive (script from document).

        Args:
            arc_engine_id: Engine ID ('arc_script', 'arc_iframe', 'arc_bat')
            arc_name: Path to script file
            fname_in_arc: Name of script to extract

        Returns:
            Extracted script data, or None on error
        """
        if arc_engine_id not in ["arc_script", "arc_iframe", "arc_bat"]:
            return None

        try:
            with open(arc_name, "rb") as fp:
                buf = fp.read()

            if arc_engine_id in ["arc_script", "arc_iframe"]:
                buf = buf.decode("latin-1")
                if obj := self.p_script_in_html.search(buf):
                    return obj.groups()[1].encode("latin-1")
            elif arc_engine_id == "arc_bat":
                buf = self.p_bat_cmt1.sub(b"", buf)
                data = self.p_bat_cmt2.sub(b"", buf)
                ret = KICOMAV_BAT_MAGIC + data
                return ret

        except (IOError, OSError) as e:
            logger.debug("Archive extract error for %s in %s: %s", fname_in_arc, arc_name, e)
        except Exception as e:
            logger.warning("Unexpected error extracting %s from %s: %s", fname_in_arc, arc_name, e)

        return None

    def mkarc(self, arc_engine_id, arc_name, file_infos):
        """Create an archive.

        Args:
            arc_engine_id: Engine ID ('arc_script', 'arc_iframe')
            arc_name: Path to archive file
            file_infos: List of file info structures

        Returns:
            True if successful, False otherwise
        """
        if arc_engine_id not in ["arc_script", "arc_iframe"]:
            return False

        try:
            with open(arc_name, "rb") as fp:
                buf = fp.read()
                buf = buf.decode("latin-1")

            if obj := self.p_script_in_html.search(buf):
                return self.update_html_scripts(obj, buf, file_infos, arc_name)

        except (IOError, OSError) as e:
            logger.error("Archive creation IO error for %s: %s", arc_name, e)
        except Exception as e:
            logger.error("Unexpected error creating archive %s: %s", arc_name, e)

        return False

    def update_html_scripts(self, obj, buf, file_infos, arc_name):
        """Update HTML scripts in the archive."""
        t = obj.group()
        pos = obj.span()

        all_script_info = [["script", pos, t]]
        org_buf = buf

        for idx, file_info in enumerate(file_infos):
            rname = file_info.get_filename()
            with contextlib.suppress(IOError):
                if os.path.exists(rname):
                    with open(rname, "rb") as fp:
                        new_buf = fp.read()

                        if len(all_script_info[idx][2]) < len(new_buf):
                            return False

                        new_buf += b" " * (len(all_script_info[idx][2]) - len(new_buf))
                        all_script_info[idx][2] = new_buf
                else:
                    new_buf = b" " * len(all_script_info[idx][2])
                    all_script_info[idx][2] = new_buf

        with open(arc_name, "wb") as fp:
            start_pos = 0
            for script_info in all_script_info:
                pos = script_info[1]
                buf_part = org_buf[start_pos : pos[0]]
                fp.write(buf_part.encode("latin-1"))
                fp.write(script_info[2])
                start_pos = pos[1]
            fp.write(org_buf[start_pos:].encode("latin-1"))

        return True

    def arcclose(self):
        """Close all open archive handles."""
        pass



# -*- coding:utf-8 -*-
# Author: Kei Choi(hanul93@gmail.com)
# Reference : https://github.com/unixfreak0037/officeparser/blob/master/officeparser.py (696~)

"""
Ole10Native Archive Engine Plugin

This plugin handles Ole10Native stream format for scanning and manipulation.
"""

import contextlib
import logging
import os
import struct

from kicomav.plugins import kernel
from kicomav.plugins import kavutil
from kicomav.kavcore import k2security
from kicomav.kavcore.plugin_base import ArchivePluginBase

# Module logger
logger = logging.getLogger(__name__)

MAX_PATH = 512


def find_null_bytes(byte_data):
    return [i for i, b in enumerate(byte_data) if b == 0]


def analysis_ole10native(mm, verbose=False):
    """Analyze Ole10Native stream format.

    Args:
        mm: File data (memory mapped)
        verbose: Debug mode flag

    Returns:
        Dictionary with format info, or None if not recognized
    """
    fileformat = {}

    try:
        if mm[:2] == b"\x02\x00":
            size = len(mm)
            fileformat["size"] = size

            index = find_null_bytes(mm[2:])
            fileformat["label"] = mm[2 : 2 + index[0]].decode("utf-8")

            off = 2 + len(fileformat["label"]) + 1
            index = find_null_bytes(mm[off:])
            fileformat["fname"] = mm[off : off + index[0]].decode("utf-8")

            off += len(fileformat["fname"]) + 1
            off += 2  # flag

            unknown_size = int(mm[off])
            off += 1 + unknown_size + 2

            index = find_null_bytes(mm[off:])
            fileformat["command"] = mm[off : off + index[0]].decode("utf-8")

            off += len(fileformat["command"]) + 1

            data_size = kavutil.get_uint32(mm, off)

            fileformat["data_off"] = off + 4
            fileformat["data_size"] = data_size

            if len(mm) < off + data_size:  # Error
                raise ValueError

            if verbose:
                print()
                kavutil.vprint("Ole10Native Stream")
                kavutil.vprint(None, "Size", f"0x{size:08X}")
                kavutil.vprint(None, "Label", fileformat["label"])
                kavutil.vprint(None, "File Name", fileformat["fname"])
                kavutil.vprint(None, "Command Line", fileformat["command"])
                kavutil.vprint(None, "Data Offset", f"0x{off + 4:08X}")
                kavutil.vprint(None, "Data Size", f"0x{data_size:08X}")

                print()
                kavutil.vprint("Data Dump")
                print()
                kavutil.HexDump().Buffer(mm[:], off + 4, 0x80)
                print()

            return fileformat
    except ValueError:
        pass
    except struct.error:
        pass

    return None


# -------------------------------------------------------------------------
# KavMain class
# -------------------------------------------------------------------------
class KavMain(ArchivePluginBase):
    """Ole10Native archive handler plugin.

    This plugin provides functionality for:
    - Detecting Ole10Native stream format
    - Extracting embedded files from Ole10Native streams
    """

    def __init__(self):
        """Initialize the Ole10Native plugin."""
        super().__init__(
            author="Kei Choi",
            version="1.0",
            title="Ole10Native Engine",
            kmd_name="olenative",
        )

    def getinfo(self):
        """Get plugin information.

        Returns:
            Dictionary containing plugin metadata
        """
        info = super().getinfo()
        info["make_arc_type"] = kernel.MASTER_PACK
        return info

    def format(self, filehandle, filename, filename_ex):
        """Analyze and detect Ole10Native format.

        Args:
            filehandle: File data (memory mapped)
            filename: Path to file
            filename_ex: Extended filename info

        Returns:
            Dictionary with format info, or empty dict if not recognized
        """
        ret = {}

        try:
            mm = filehandle

            if mm[:2] == b"\x02\x00" and filename_ex.find("_Ole10Native") != -1:
                if fileformat := analysis_ole10native(mm, self.verbose):
                    ret = {"ff_ole10native": fileformat}

        except (IOError, OSError) as e:
            logger.debug("Format detection IO error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error in format detection for %s: %s", filename, e)

        return ret

    def arclist(self, filename, fileformat, password=None):
        """List files in the archive.

        Args:
            filename: Path to archive file
            fileformat: Format info from format() method

        Returns:
            List of [engine_id, filename] pairs
        """
        file_scan_list = []

        if "ff_ole10native" not in fileformat:
            return file_scan_list

        try:
            fformat = fileformat["ff_ole10native"]
            name = fformat["label"]  # File name hidden inside OLE

            # CWE-22: Path traversal prevention
            if not k2security.is_safe_archive_member(name):
                logger.warning("Unsafe archive member rejected: %s in %s", name, filename)
                return file_scan_list

            off = fformat["data_off"]
            data_size = fformat["data_size"]

            arc_name = f"arc_ole10native:{off}:{data_size}"
            file_scan_list.append([arc_name, name])

        except (IOError, OSError) as e:
            logger.debug("Archive list IO error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error listing archive %s: %s", filename, e)

        return file_scan_list

    def unarc(self, arc_engine_id, arc_name, fname_in_arc):
        """Extract a file from the archive.

        Args:
            arc_engine_id: Engine ID ('arc_ole10native:offset:size')
            arc_name: Path to archive file
            fname_in_arc: Name of file to extract

        Returns:
            Extracted file data, or None on error
        """
        # CWE-22: Path traversal prevention
        if not k2security.is_safe_archive_member(fname_in_arc):
            logger.warning("Unsafe archive member rejected: %s in %s", fname_in_arc, arc_name)
            return None

        if arc_engine_id.find("arc_ole10native:") == -1:
            return None

        try:
            val = arc_engine_id.split(":")
            off = int(val[1])
            size = int(val[2])

            with open(arc_name, "rb") as f:
                buf = f.read()
                return buf[off : off + size]

        except (IOError, OSError) as e:
            logger.debug("Archive extract IO error for %s in %s: %s", fname_in_arc, arc_name, e)
        except Exception as e:
            logger.warning("Unexpected error extracting %s from %s: %s", fname_in_arc, arc_name, e)

        return None

    def arcclose(self):
        """Close all open archive handles."""
        pass  # No persistent handles to close

    def mkarc(self, arc_engine_id, arc_name, file_infos):
        """Create an Ole10Native archive.

        Args:
            arc_engine_id: Engine ID ('arc_ole10native:offset:size')
            arc_name: Path to archive file
            file_infos: List of file info structures

        Returns:
            True if successful, False otherwise
        """
        if arc_engine_id.find("arc_ole10native:") == -1:
            return False

        try:
            val = arc_engine_id.split(":")
            off = int(val[1])
            size = int(val[2])

            with open(arc_name, "rb") as fp:
                ole10native_data = fp.read()

            file_info = file_infos[0]
            rname = file_info.get_filename()

            with open(rname, "rb") as fp:
                buf = fp.read()

                new_data = ole10native_data[: off - 4]  # Delete original data
                new_data += struct.pack("<L", len(buf)) + buf  # Replace with new data
                new_data += ole10native_data[off + size :]

                with open(arc_name, "wb") as out_fp:
                    out_fp.write(new_data)

            return True

        except (IOError, OSError) as e:
            logger.error("Archive creation IO error for %s: %s", arc_name, e)
        except Exception as e:
            logger.error("Unexpected error creating archive %s: %s", arc_name, e)

        return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
KICOMAV Plugin Base Classes

This module provides abstract base classes for all KICOMAV plugins,
implementing common functionality and enforcing consistent interfaces.
"""

import os
import logging
import pathlib
from abc import ABC, abstractmethod
from enum import IntEnum
from typing import Dict, Any, List, Optional, Tuple, BinaryIO, Union

# Module logger
logger = logging.getLogger(__name__)

from . import k2security
from . import k2const


# Default maximum scan size (100MB)
# Can be overridden via init() or set_max_scan_size()
DEFAULT_MAX_SCAN_SIZE = 100 * 1024 * 1024

# Special value to disable size limit
UNLIMITED_SCAN_SIZE = 0


# -------------------------------------------------------------------------
# Scan Result Constants
# -------------------------------------------------------------------------
class ScanResultCode(IntEnum):
    """Scan result codes for threat identification."""

    CLEAN = 0  # No threat detected (normal scan completion)
    INFECTED = 1  # Threat detected via signature
    HEURISTIC = 2  # Threat detected via heuristic analysis
    ERROR = -1  # Scan error occurred
    SKIPPED = -2  # File was skipped (too large, unsupported, etc.)


# Threat ID constants
THREAT_ID_DETECTED = 0  # Normal detection
THREAT_ID_NOT_FOUND = -1  # No threat found or error

# Empty scan result tuple (no threat found)
SCAN_RESULT_CLEAN = (False, "", THREAT_ID_NOT_FOUND, "")


class PluginBase(ABC):
    """Base class for all KICOMAV plugins.

    This abstract base class provides common initialization, cleanup,
    and information retrieval functionality for all plugin types.
    """

    def __init__(self, author: str = "Kei Choi", version: str = "1.0", title: str = "", kmd_name: str = ""):
        """Initialize the plugin with metadata.

        Args:
            author: Plugin author name
            version: Plugin version string
            title: Human-readable plugin title
            kmd_name: Internal plugin identifier
        """
        self.author = author
        self.version = version
        self.title = title
        self.kmd_name = kmd_name
        self.verbose = False
        self.rules_paths = {}  # {"system": "/path", "user": "/path" or None}
        self._max_scan_size = DEFAULT_MAX_SCAN_SIZE

    def init(
        self,
        rules_paths: Optional[Dict[str, Optional[str]]] = None,
        verbose: bool = False,
        max_scan_size: Optional[int] = None,
    ) -> int:
        """Initialize the plugin engine.

        Args:
            rules_paths: Dictionary with rule paths {"system": "/path", "user": "/path"}
            verbose: Enable verbose output
            max_scan_size: Maximum file size to scan in bytes.
                          None = use default (100MB)
                          0 = unlimited

        Returns:
            0 for success, non-zero for failure
        """
        # Validate rules_paths is dict (strict enforcement)
        if rules_paths is not None and not isinstance(rules_paths, dict):
            raise TypeError(f"rules_paths must be a dict, got {type(rules_paths).__name__}")

        self.rules_paths = rules_paths or {"system": None, "user": None}
        self.verbose = verbose

        if max_scan_size is not None:
            self._max_scan_size = max_scan_size

        # Call subclass-specific initialization
        return self._custom_init()

    def _get_rule_path(self, rule_type: str) -> List[pathlib.Path]:
        """Get rule paths for a specific rule type.

        Each plugin can use this to find its rule directory.
        For example, YARA plugin uses 'yara', KicomAV pattern uses 'kicomav'.

        Args:
            rule_type: Subfolder name ('yara', 'kicomav', etc.)

        Returns:
            List of valid Path objects for the rule type
        """
        paths = []
        for key in ["system", "user"]:
            base = self.rules_paths.get(key)
            if base:
                path = pathlib.Path(base) / rule_type
                if path.exists() and path.is_dir():
                    paths.append(path)
        return paths

    @property
    def max_scan_size(self) -> int:
        """Get maximum scan size in bytes.

        Returns:
            Maximum scan size (0 = unlimited)
        """
        return self._max_scan_size

    @max_scan_size.setter
    def max_scan_size(self, value: int):
        """Set maximum scan size.

        Args:
            value: Maximum size in bytes (0 = unlimited)
        """
        if value < 0:
            raise ValueError("max_scan_size must be non-negative")
        self._max_scan_size = value

    def _custom_init(self) -> int:
        """Subclass-specific initialization logic.

        Override this method in subclasses to add custom initialization.

        Returns:
            0 for success, non-zero for failure
        """
        return 0

    def uninit(self) -> int:
        """Terminate the plugin engine and cleanup resources.

        Returns:
            0 for success, non-zero for failure
        """
        self._cleanup_resources()
        return self._custom_uninit()

    def _custom_uninit(self) -> int:
        """Subclass-specific cleanup logic.

        Override this method in subclasses to add custom cleanup.

        Returns:
            0 for success, non-zero for failure
        """
        return 0

    def _cleanup_resources(self):
        """Common resource cleanup logic."""
        self.verbose = False

    def _validate_path_input(self, path: str, context: str = "path") -> bool:
        """Validate path input for basic security checks.

        Performs lightweight validation without requiring base_dir.
        For full path traversal protection, use k2security.validate_path().

        Args:
            path: Path string to validate
            context: Context for logging (e.g., "filename", "archive")

        Returns:
            True if path is safe, False otherwise
        """
        if not path:
            return True  # Empty path is handled elsewhere

        # Check for null bytes
        if "\0" in path:
            logger.warning("Null byte detected in %s: %s", context, repr(path))
            return False

        # Check for path traversal patterns
        path_parts = pathlib.Path(path).parts
        if ".." in path_parts:
            logger.warning("Path traversal detected in %s: %s", context, path)
            return False

        return True

    def _is_safe_archive_member(self, member_name: str) -> bool:
        """Check if archive member name is safe.

        Args:
            member_name: Name of the archive member

        Returns:
            True if safe, False otherwise
        """
        return k2security.is_safe_archive_member(member_name)

    def getinfo(self) -> Dict[str, Any]:
        """Get plugin information.

        Returns:
            Dictionary containing plugin metadata
        """
        return {"author": self.author, "version": self.version, "title": self.title, "kmd_name": self.kmd_name}

    def listvirus(self, *args, **kwargs) -> List[str]:
        """Return list of detectable viruses/threats.

        Override in malware detector plugins.
        Archive and format plugins may leave this unimplemented.

        Returns:
            List of virus names this plugin can detect
        """
        return []

    def scan(self, filehandle: Union[bytes, BinaryIO], filename: str, *args, **kwargs) -> Tuple[bool, str, int, str]:
        """Scan a file for threats.

        Override in malware detector plugins.
        Archive and format plugins may leave this unimplemented.

        Args:
            filehandle: File data or handle to scan
            filename: Name of the file being scanned
            *args: Additional positional arguments
            **kwargs: Additional keyword arguments

        Returns:
            Tuple of (threat_found, threat_name, threat_id, plugin_name)
        """
        return (False, "", -1, "")


class ArchivePluginBase(PluginBase):
    """Base class for archive processing plugins.

    Provides common functionality for handling various archive formats
    like ZIP, RAR, CAB, TAR, etc.

    Subclasses can either:
    1. Override high-level methods (format, arclist, unarc, mkarc) directly
       for complex multi-format plugins
    2. Implement helper methods (_get_signature, _parse_archive_format, etc.)
       for simple single-format plugins
    """

    def __init__(self, *args, **kwargs):
        """Initialize archive plugin."""
        super().__init__(*args, **kwargs)
        self.handle = {}  # Dict for multi-format support: {filename: handle}

    def format(self, filehandle: bytes, filename: str, filename_ex: str) -> Optional[Dict[str, Any]]:
        """Verify and parse archive format.

        Override this method for multi-format support or custom detection.

        Args:
            filehandle: File data to check
            filename: Original filename
            filename_ex: Extended filename information

        Returns:
            Dictionary with format information, None if not recognized
        """
        try:
            mm = filehandle
            signature = self._get_signature()

            if signature and len(mm) >= len(signature) and mm[: len(signature)] == signature:
                return self._parse_archive_format(mm, filename, filename_ex)
        except (IOError, OSError) as e:
            logger.debug("Archive format IO error for %s: %s", filename, e)
            raise k2const.PluginUnexpectedError(str(e)) from e
        except (ValueError, TypeError) as e:
            logger.debug("Archive format parse error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error in archive format check for %s: %s", filename, e)
            raise k2const.PluginUnexpectedError(str(e)) from e

        return None

    def _get_signature(self) -> bytes:
        """Get the archive format signature.

        Override for single-format plugins.
        Return empty bytes if using custom format() implementation.

        Returns:
            Bytes representing the archive signature
        """
        return b""

    def _parse_archive_format(self, mm: bytes, filename: str, filename_ex: str) -> Dict[str, Any]:
        """Parse archive format and extract metadata.

        Override for single-format plugins.

        Args:
            mm: Memory-mapped file data
            filename: Original filename
            filename_ex: Extended filename information

        Returns:
            Dictionary containing archive metadata
        """
        return {}

    def arclist(self, filename: str, fileformat: Dict[str, Any], password: Optional[str] = None) -> List[Any]:
        """List files in the archive.

        Override this method for custom archive listing behavior.
        Return format can be List[Dict] or List[List] depending on needs.

        Args:
            filename: Path to archive file
            fileformat: Format information from format() method
            password: Optional password for encrypted archives

        Returns:
            List of file information (format depends on implementation)
        """
        file_scan_list = []

        try:
            with self._open_archive(filename, fileformat) as archive:
                for member in self._get_members(archive):
                    file_info = self._extract_member_info(member)
                    if file_info:
                        file_scan_list.append(file_info)
        except (IOError, OSError) as e:
            logger.debug("Archive list IO error for %s: %s", filename, e)
            raise k2const.PluginUnexpectedError(str(e)) from e
        except (ValueError, KeyError) as e:
            logger.debug("Archive list parse error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error listing archive %s: %s", filename, e)
            raise k2const.PluginUnexpectedError(str(e)) from e

        return file_scan_list

    def _open_archive(self, filename: str, fileformat: Dict[str, Any]):
        """Open archive file as context manager.

        Override for single-format plugins.

        Args:
            filename: Path to archive file
            fileformat: Format information

        Returns:
            Archive object (context manager)
        """
        raise NotImplementedError("Subclass must implement _open_archive or override arclist")

    def _get_members(self, archive) -> List[Any]:
        """Get list of archive members.

        Override for single-format plugins.

        Args:
            archive: Opened archive object

        Returns:
            List of archive members
        """
        return []

    def _extract_member_info(self, member) -> Any:
        """Extract information about an archive member.

        Override for single-format plugins.

        Args:
            member: Archive member object

        Returns:
            Member information (format depends on implementation)
        """
        return None

    def unarc(self, arc_engine_id: str, arc_name: str, fname_in_arc: str) -> Optional[bytes]:
        """Extract a file from the archive.

        Override this method for custom extraction behavior.

        Args:
            arc_engine_id: Engine ID for this archive (e.g., 'arc_zip')
            arc_name: Path to archive file
            fname_in_arc: Name of file to extract

        Returns:
            Extracted file data or None on error
        """
        # Validate archive member name for path traversal
        if not self._is_safe_archive_member(fname_in_arc):
            logger.warning("Unsafe archive member rejected: %s in %s", fname_in_arc, arc_name)
            return None

        try:
            return self._extract_file(arc_engine_id, arc_name, fname_in_arc)
        except (IOError, OSError) as e:
            logger.debug("Archive extract IO error for %s in %s: %s", fname_in_arc, arc_name, e)
            raise k2const.PluginUnexpectedError(str(e)) from e
        except (ValueError, KeyError, IndexError) as e:
            logger.debug("Archive extract parse error for %s in %s: %s", fname_in_arc, arc_name, e)
        except Exception as e:
            logger.warning("Unexpected error extracting %s from %s: %s", fname_in_arc, arc_name, e)
            raise k2const.PluginUnexpectedError(str(e)) from e

        return None

    def _extract_file(self, arc_engine_id: str, arc_name: str, fname_in_arc: str) -> Optional[bytes]:
        """Extract specific file from archive.

        Override for single-format plugins.

        Args:
            arc_engine_id: Engine ID for this archive
            arc_name: Path to archive file
            fname_in_arc: Name of file to extract

        Returns:
            Extracted file data
        """
        return None

    def arcclose(self):
        """Close all open archive handles.

        Override if using different handle management.
        """
        for fname in list(self.handle.keys()):
            try:
                handle = self.handle[fname]
                if hasattr(handle, "close"):
                    handle.close()
            except (IOError, OSError) as e:
                logger.debug("Archive close IO error for %s: %s", fname, e)
            except Exception as e:
                logger.debug("Archive close error for %s: %s", fname, e)
            finally:
                self.handle.pop(fname, None)

    def _cleanup_resources(self):
        """Clean up archive-specific resources."""
        super()._cleanup_resources()
        self.arcclose()

    def mkarc(self, arc_engine_id: str, arc_name: str, file_infos: List[Any]) -> bool:
        """Create an archive with specified files.

        Override this method for custom archive creation.

        Args:
            arc_engine_id: Engine ID for archive type
            arc_name: Path for new archive
            file_infos: List of file information

        Returns:
            True if successful, False otherwise
        """
        try:
            return self._create_archive(arc_engine_id, arc_name, file_infos)
        except (IOError, OSError) as e:
            logger.error("Archive creation IO error for %s: %s", arc_name, e)
            raise k2const.PluginUnexpectedError(str(e)) from e
        except (ValueError, TypeError) as e:
            logger.error("Archive creation data error for %s: %s", arc_name, e)
        except Exception as e:
            logger.error("Unexpected error creating archive %s: %s", arc_name, e)
            raise k2const.PluginUnexpectedError(str(e)) from e

        return False

    def _create_archive(self, arc_engine_id: str, arc_name: str, file_infos: List[Any]) -> bool:
        """Create archive implementation.

        Override in subclasses that support archive creation.

        Args:
            arc_engine_id: Engine ID for archive type
            arc_name: Path for new archive
            file_infos: List of file information

        Returns:
            True if successful, False otherwise
        """
        return False


class MalwareDetectorBase(PluginBase):
    """Base class for malware detection plugins.

    Provides common functionality for signature-based and
    heuristic malware detection.
    """

    def __init__(self, *args, **kwargs):
        """Initialize malware detector."""
        super().__init__(*args, **kwargs)
        self.virus_names = []
        self.virus_info = []
        self.virus_patterns = {}

    def _custom_init(self) -> int:
        """Load virus database during initialization.

        Returns:
            0 for success, non-zero for failure
        """
        return self._load_virus_database()

    @abstractmethod
    def _load_virus_database(self) -> int:
        """Load virus signatures and patterns.

        Returns:
            0 for success, non-zero for failure
        """
        pass

    def listvirus(self, *args, **kwargs) -> List[str]:
        """Get list of detectable viruses.

        Returns:
            List of virus names
        """
        return self.virus_names

    def scan(self, filehandle: Union[bytes, BinaryIO], filename: str, *args, **kwargs) -> Tuple[bool, str, int, str]:
        """Scan file for malware.

        Args:
            filehandle: File data or handle to scan
            filename: Name of the file being scanned
            *args: Additional positional arguments
            **kwargs: Additional keyword arguments

        Returns:
            Tuple of (threat_found, threat_name, threat_id, plugin_name)
        """
        try:
            # Get file data with size limit
            if hasattr(filehandle, "read"):
                if self._max_scan_size > 0:
                    mm = filehandle.read(self._max_scan_size)
                    # Check if file was truncated
                    if hasattr(filehandle, "seek") and hasattr(filehandle, "tell"):
                        current_pos = filehandle.tell()
                        filehandle.seek(0, 2)  # Seek to end
                        file_size = filehandle.tell()
                        filehandle.seek(current_pos)  # Restore position
                        if file_size > self._max_scan_size:
                            logger.info(
                                "File %s truncated for scan: %d > %d bytes", filename, file_size, self._max_scan_size
                            )
                else:
                    mm = filehandle.read()
            else:
                mm = filehandle
                # Apply size limit to bytes data
                if self._max_scan_size > 0 and len(mm) > self._max_scan_size:
                    logger.info("Data %s truncated for scan: %d > %d bytes", filename, len(mm), self._max_scan_size)
                    mm = mm[: self._max_scan_size]

            # Signature-based scanning
            for vname, pattern in self.virus_patterns.items():
                if self._match_pattern(mm, pattern):
                    return True, vname, THREAT_ID_DETECTED, self.kmd_name

            # Heuristic scanning
            heuristic_result = self._heuristic_scan(mm, filename)
            if heuristic_result:
                return True, "Heuristic.Suspicious", THREAT_ID_DETECTED, self.kmd_name

        except (IOError, OSError) as e:
            logger.debug("Scan IO error for %s: %s", filename, e)
            raise k2const.PluginUnexpectedError(str(e)) from e
        except (ValueError, TypeError) as e:
            logger.debug("Scan data error for %s: %s", filename, e)
        except MemoryError as e:
            logger.error("Scan memory error for %s (file too large?): %s", filename, e)
            raise k2const.PluginUnexpectedError(str(e)) from e
        except Exception as e:
            logger.warning("Unexpected error scanning %s: %s", filename, e)
            raise k2const.PluginUnexpectedError(str(e)) from e

        return SCAN_RESULT_CLEAN

    def _match_pattern(self, data: bytes, pattern: Any) -> bool:
        """Match data against a virus pattern.

        Override in subclasses that use the base scan() method with
        virus_patterns dictionary. Not needed if scan() is overridden.

        Args:
            data: File data to check
            pattern: Virus pattern to match

        Returns:
            True if pattern matches, False otherwise
        """
        return False

    def _heuristic_scan(self, data: bytes, filename: str) -> bool:
        """Perform heuristic scanning.

        Override in subclasses to add heuristic detection.

        Args:
            data: File data to analyze
            filename: Name of the file

        Returns:
            True if suspicious, False otherwise
        """
        return False

    def disinfect(self, filename: str, malware_id: int) -> bool:
        """Attempt to disinfect a file.

        Override in subclasses that support disinfection.

        Args:
            filename: Path to infected file
            malware_id: ID of detected malware

        Returns:
            True if disinfected, False otherwise
        """
        # Validate filename for path traversal
        if not self._validate_path_input(filename, "disinfect_filename"):
            return False

        return False

    def _cleanup_resources(self):
        """Clean up malware detector resources."""
        super()._cleanup_resources()
        # Use new assignments instead of clear() to avoid
        # affecting external references
        self.virus_names = []
        self.virus_info = []
        self.virus_patterns = {}


class FileFormatPluginBase(PluginBase):
    """Base class for file format processing plugins.

    Provides common functionality for handling specific file formats
    like PE, ELF, PDF, Office documents, etc.
    """

    def __init__(self, *args, **kwargs):
        """Initialize file format plugin."""
        super().__init__(*args, **kwargs)
        self.signatures = {}

    @abstractmethod
    def format(self, filehandle: bytes, filename: str, filename_ex: str) -> Dict[str, Any]:
        """Verify and parse file format.

        Args:
            filehandle: File data to check
            filename: Original filename
            filename_ex: Extended filename information

        Returns:
            Dictionary with format information, empty if not recognized
        """
        pass

    def get_format_name(self) -> str:
        """Get the format name handled by this plugin.

        Returns:
            Format name string
        """
        return self.kmd_name

    def _check_signature(self, data: bytes, offset: int = 0) -> Optional[str]:
        """Check data against known signatures.

        Args:
            data: Data to check
            offset: Offset to start checking from

        Returns:
            Signature name if matched, None otherwise
        """
        for sig_name, sig_bytes in self.signatures.items():
            sig_len = len(sig_bytes)
            if len(data) >= offset + sig_len:
                if data[offset : offset + sig_len] == sig_bytes:
                    return sig_name
        return None

    def extract_embedded(self, filehandle: bytes, filename: str) -> List[Tuple[str, bytes]]:
        """Extract embedded files or objects.

        Override in subclasses that support extraction.

        Args:
            filehandle: File data
            filename: Original filename

        Returns:
            List of tuples (name, data) for embedded objects
        """
        return []

    def repair(self, filename: str, fileformat: Dict[str, Any]) -> bool:
        """Attempt to repair a corrupted file.

        Override in subclasses that support file repair.

        Args:
            filename: Path to file to repair
            fileformat: Format information

        Returns:
            True if repaired, False otherwise
        """
        # Validate filename for path traversal
        if not self._validate_path_input(filename, "repair_filename"):
            return False

        return False



# -*- coding:utf-8 -*-
# Author: Kei Choi(hanul93@gmail.com)

"""
Virus Engine Plugin

This plugin handles virus type malware detection using signature patterns.
"""

import contextlib
import logging
import os

from kicomav.plugins import cryptolib
from kicomav.plugins import kavutil
from kicomav.plugins import kernel
from kicomav.kavcore import k2security
from kicomav.kavcore.plugin_base import MalwareDetectorBase

# Module logger
logger = logging.getLogger(__name__)


# -------------------------------------------------------------------------
# Structure of the signature file
# -------------------------------------------------------------------------
# Flag + Word(Offset for the corresponding Flag is always at the 0th position of the 2Byte)
# Flag
# 0 : Start of the file
# 1 : Execution position (DOS-EP)
# 2 : Execution position (PE-EP)
# 3 : Start of each section (PE, ELF, etc.)
# 4 : Start of Attach
# Checksum1 : Flag, Offset, Length, CRC32
# Checksum2 : Flag, Offset, Length, CRC32
# MalwareName
# -------------------------------------------------------------------------
# Example:  0000 F8A8:02, 0000, 0000, XXXXXXXX:02, 0000, 0000, XXXXXXXX:MalwareName
# -------------------------------------------------------------------------


# -------------------------------------------------------------------------
# Pre-create patterns for specific sizes in the given buffer.
# -------------------------------------------------------------------------
def gen_checksums(buf):
    patterns = []

    # First 10 are 6, 7, 8, 9 ... 0xF
    patterns.extend(int(gen_checksum(buf, 0, i), 16) for i in range(1, 0x10))

    # The remaining 15 are 0x10, 0x18, 0x20 ... 0x80
    patterns.extend(int(gen_checksum(buf, 0, i), 16) for i in range(0x10, 0x88, 8))
    return patterns


def gen_checksum(buf, off, size):
    return cryptolib.crc32(buf[off : off + size])


# -------------------------------------------------------------------------
# KavMain Class
# -------------------------------------------------------------------------
class KavMain(MalwareDetectorBase):
    """Virus malware detector plugin.

    This plugin provides functionality for:
    - Detecting virus type malware using signature patterns
    - CRC32 checksum based detection
    """

    def __init__(self):
        """Initialize the Virus Engine plugin."""
        super().__init__(
            author="Kei Choi",
            version="1.0",
            title="Virus Engine",
            kmd_name="ve",
        )
        self.flags_off = {}

    def _load_virus_database(self) -> int:
        """Load virus patterns.

        Returns:
            0 for success
        """
        self.flags_off = {}
        return 0

    def getinfo(self):
        """Get plugin information.

        Returns:
            Dictionary containing plugin metadata
        """
        info = super().getinfo()
        info["sig_num"] = kavutil.handle_pattern_vdb.get_sig_num("ve") + 2
        return info

    def listvirus(self):
        """Get list of detectable viruses.

        Returns:
            List of virus names
        """
        vlist = kavutil.handle_pattern_vdb.get_sig_vlist("ve")

        vlists = ["Virus.Win32.Small.a", "Virus.Win32.SuperThreat.b"]
        if vlist:
            vlists.extend(kavutil.normal_vname(vname) for vname in vlist)
        vlists.sort()
        return vlists

    def scan(self, filehandle, filename, fileformat, filename_ex):
        """Scan for malware.

        Args:
            filehandle: File data (memory mapped)
            filename: Path to file
            fileformat: Format info from format() method
            filename_ex: Extended filename info

        Returns:
            Tuple of (found, malware_name, malware_id, result)
        """
        try:
            self.flags_off = {}
            mm = filehandle

            # Scan for Virus.Win32.Small.a
            ret, vname = self.__scan_virus_win32_small_a(filehandle, fileformat)
            if ret:
                return True, vname, 0, kernel.INFECTED

            flags = [[int(f"0000{mm[:2].hex()}", 16), gen_checksums(mm[:0x80])]]
            self.flags_off[0] = [0]

            # Is there a PE format in the pre-analyzed file format?
            if "ff_pe" in fileformat:
                self.process_pe_flags_and_checksums(fileformat, flags, mm)

            # Is there an attach area?
            if "ff_attach" in fileformat:
                size = fileformat["ff_attach"]["Attached_Size"]
                if size > 0x80:
                    # Flag - 4 : Attach area
                    pos = fileformat["ff_attach"]["Attached_Pos"]
                    flags.append(
                        [
                            int(f"0004{mm[pos : pos + 2].hex()}", 16),
                            gen_checksums(mm[pos : pos + 0x80]),
                        ]
                    )
                    self.flags_off[4] = [pos]

            cs_size = [
                1,
                2,
                3,
                4,
                5,
                6,
                7,
                8,
                9,
                0xA,
                0xB,
                0xC,
                0xD,
                0xE,
                0xF,
                0x10,
                0x18,
                0x20,
                0x28,
                0x30,
                0x38,
                0x40,
                0x48,
                0x50,
                0x58,
                0x60,
                0x68,
                0x70,
                0x78,
                0x80,
            ]

            if self.verbose:
                self.write_debug_vdb_file(filename, mm, flags, cs_size)

            for flag in flags:
                if p1 := kavutil.handle_pattern_vdb.match_size("ve", flag[0]):
                    for ve_id in p1.keys():
                        for idx in p1[ve_id]:
                            cs1 = kavutil.handle_pattern_vdb.get_cs1(ve_id, idx)

                            cs1_flag = cs1[0]
                            cs1_off = cs1[1]
                            cs1_size = cs1[2]
                            cs1_crc = cs1[3]

                            if flag[0] >> 16 == cs1_flag and cs1_off == 0 and cs1_size in cs_size:
                                i = cs_size.index(cs1_size)

                                if cs1_crc == flag[1][i]:  # Is the first pattern the same?
                                    if vname := self.__scan_cs2(mm, ve_id, idx):
                                        return True, vname, 0, kernel.INFECTED
                            else:
                                buf = self.__get_data_crc32(mm, cs1_flag, cs1_off, cs1_size)
                                if cs1_crc == int(gen_checksum(mm, cs1_off, cs1_size), 16):
                                    if vname := self.__scan_cs2(mm, ve_id, idx):
                                        return True, vname, 0, kernel.INFECTED

        except (IOError, OSError) as e:
            logger.debug("Scan IO error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error scanning %s: %s", filename, e)

        return False, "", -1, kernel.NOT_FOUND

    def process_pe_flags_and_checksums(self, fileformat, flags, mm):
        """Process PE format flags and checksums."""
        # Flag - 2 : PE EP
        ff = fileformat["ff_pe"]
        ep_off = ff["pe"]["EntryPointRaw"]
        flags.append(
            [
                int(f"0002{mm[ep_off : ep_off + 2].hex()}", 16),
                gen_checksums(mm[ep_off : ep_off + 0x80]),
            ]
        )
        self.flags_off[2] = [ep_off]

        # Flag - 3 : Start of each section
        flag3_off = []
        for section in ff["pe"]["Sections"]:
            fsize = section["SizeRawData"]
            foff = section["PointerRawData"]
            flags.append(
                [
                    int(f"0003{mm[foff : foff + 2].hex()}", 16),
                    gen_checksums(mm[foff : foff + 0x80]),
                ]
            )
            flag3_off.append(foff)
        self.flags_off[3] = flag3_off

    def write_debug_vdb_file(self, filename, mm, flags, cs_size):
        """Write debug VDB file in verbose mode."""
        print("-" * 79)
        kavutil.vprint("Engine")
        kavutil.vprint(None, "Engine", "ve")
        kavutil.vprint(None, "File name", os.path.split(filename)[-1])
        kavutil.vprint(None, "MD5", cryptolib.md5(mm[:]))

        print()
        kavutil.vprint("VE")
        vdb_name = f"{os.path.split(filename)[-1]}.vdb"
        kavutil.vprint(None, "VDB File name", vdb_name)

        with open(vdb_name, "w") as fp:
            for flag in flags:
                msg = "Flag : %08x\n" % flag[0]
                fp.write(msg)

                for i, cs in enumerate(flag[1]):
                    msg = "CS = %02x : %08x\n" % (cs_size[i], int(cs))
                    fp.write(msg)
                fp.write("\n")

    def disinfect(self, filename, malware_id):
        """Disinfect malware.

        Args:
            filename: Path to infected file
            malware_id: Malware ID to disinfect

        Returns:
            True if successful, False otherwise
        """
        try:
            if malware_id == 0:
                filename_dir = os.path.dirname(filename) or os.getcwd()
                k2security.safe_remove_file(filename, filename_dir)
                return True

        except (IOError, OSError, k2security.SecurityError) as e:
            logger.debug("Disinfect error for %s: %s", filename, e)
        except Exception as e:
            logger.warning("Unexpected error disinfecting %s: %s", filename, e)

        return False

    def __get_data_crc32(self, buf, flag, off, size):
        """Get the crc32 of a specific position."""
        base_offs = self.flags_off.get(flag, [])
        return [int(gen_checksum(buf, base_off + off, size), 16) for base_off in base_offs]

    def __scan_cs2(self, mm, ve_id, idx):
        """Scan for the second pattern."""
        cs2 = kavutil.handle_pattern_vdb.get_cs2(ve_id, idx)
        cs2_flag = cs2[0]
        cs2_off = cs2[1]
        cs2_size = cs2[2]
        cs2_crc = cs2[3]
        vname_id = cs2[4]

        crc32s = self.__get_data_crc32(mm, cs2_flag, cs2_off, cs2_size)
        if cs2_crc in crc32s:  # Pattern match
            if vname := kavutil.handle_pattern_vdb.get_vname(ve_id, vname_id):
                return kavutil.normal_vname(vname)

        return None

    def __scan_virus_win32_small_a(self, mm, fileformat):
        """Scan for Virus.Win32.Small.a."""
        if "ff_pe" in fileformat:
            ff = fileformat["ff_pe"]["pe"]
            ep_off = ff["EntryPointRaw"]

            if cryptolib.crc32(mm[ep_off : ep_off + 12]) == "4d49a25f":
                v_rva = kavutil.get_uint32(mm, ep_off + 12) + 1  # Malware RVA
                v_rva -= ff["ImageBase"]

                # Check if v_rva is a value in the last section.
                sec = ff["Sections"][-1]
                if sec["RVA"] <= v_rva <= sec["RVA"] + sec["VirtualSize"]:
                    pe_file_align = ff["FileAlignment"]
                    if pe_file_align:
                        foff = (sec["PointerRawData"] // pe_file_align) * pe_file_align
                    else:
                        foff = sec["PointerRawData"]

                    v_off = v_rva - sec["RVA"] + foff

                    x = cryptolib.crc32(mm[v_off : v_off + 0x30])
                    if x == "8d964738":
                        return True, "Virus.Win32.Small.a"
                    elif x in ["00000000", "f288b395"]:
                        return True, "Virus.Win32.SuperThreat.b"

        return False, None
